/*{{{  map.c*/
#ifdef V_ADAPT

/*{{{  *** DISCUSSION -- when are 'mapp' and 'sectp' defined? ****/
/*----------------------------------------------------------------
DISCUSSION -- when are 'mapp' and 'sectp' defined?
--------------------------------------------------
In 'symtab'
~~~~~~~~~~~
'sp->mapp' (but *NOT* 'sp->sectp') is defined for:

-- all arrays (except processors and templates), even if they're
	globally replicated.
-- scalars *if and only if* they're not globally replicated, i.e. if
	they're explicitly distributed, inherited, or aligned with
	an ultimate align target that is explicitly distributed or inherited.

I.e. all data entities 'sp' have 'sp->mapp' *except* for globally
replicated scalars (i.e. ones that aren't embedded!).

Expr nodes 'ep'
~~~~~~~~~~~~~~~
'Map' and 'Sect' structs are installed in 'ep' only by:

-- attach_map (ep->mapp, mapp)			-- installs 'ep->mapp'
-- evaluate_subscrs_and_install_sect (ep,...)	-- installs 'ep->sectp'

These are called by:

-- 'transform_and_write_assignment()',
	which sets them for a non-replicated-scalar assignment var
	(and also calls 'compose_map_and_sect()') to produce equiv
	'tgt_mapp' & 'sectp' for the RHS, which are passed to the following...)

-- 'attain_mapping (ep, tgt_mapp, tgt_sectp)',
	which sets them for all non-replicated-scalar primaries on the
	RHS of an assignment;

-- 'write_remap ()' & 'write_gather ()'  -- which creates a
	'ND_TEMP_ARRAY' or 'ND_TEMP_SCALAR' and attaches 'tgt_mapp'
	& 'tgt_sectp' to it (via 'alloc_var ()' and 'alloc_mapvec ()').

What is attached to what nodes?

-- ND_TEMPARRAY, ND_WHOLE_ARRAY, ND_ARRAY_SECT	-- ep->mapp & ep->sectp;

-- ND_SIMPLE_SCALAR	-- 'ep->mapp' *if* it's embedded in the processor array.
	(But it never has 'ep->sectp', as it has no dimensions).

-- No other expr nodes have 'ep->mapp' or 'ep->sectp'.

I.e. *all* dimensioned objects have both 'ep->mapp' and 'ep->sectp'.
A non-globally-replicated simple scalar (i.e. an embedded one!)
has 'ep->mapp' only.
----------------------------------------------------------------*/
/*}}}*/

/***********************************
-- Problem with spec exprs (i.e. among decls).  Then 'sp->mapp' not
installed yet, so get errors (in fact may crash) with mapped ids,
especially array refs!

To do:
------
-- Remove MA_TYPE field from 'mapp->attrs'.
**********************************/
/*{{{  decls*/

#include "adout.h"	/* for 'mapvec' macros */
#include "expr.h"
#include "map.h"
#include "output.h"	/* for 'increase_ / reduce_indent' */
#include "strmacs.h"
#include "symbol.h"
#include "typedef.h"

Align   *inherit_alignp;	/* static 'Align' record meaning 'INHERIT' */
Map     *inherit_mapp;		/*    "   'Map'      "       "       "     */

Symbol  *default_proc_sp [MAX_RANK + 1];	/* for "$p0"--"$p7" */

bool  must_call_scatter ();
/*}}}*/

/*{{{  << HPF initialisation >>*/

/*{{{  init_HPF_records ()*/
init_HPF_records ()
/*---------------*/
/*--------------------------------------------------------------*/
/*	Called by 'main ()' to create 'inherit_alignp' and	*/
/* 'inherit_mapp' records for INHERITed dummy arguments.  These	*/
/* records have a fixed format and are statically available	*/
/* throughout the program execution.				*/
/*--------------------------------------------------------------*/
{
	int  rank;

	/*------------------------------------------------------*/
	/* create the static symbol table 'def_proc_sp []' for	*/
	/* default processor arrangements "$p0" -- "$p7".	*/
	/*------------------------------------------------------*/
	create_default_proc_sp ();

	/*------------------------*/
	/* setup 'inherit_alignp' */
	/*------------------------*/
	attach_align (inherit_alignp, create_align ());
	inherit_alignp->attrs = AL_CHECKED_TGT;

	/*----------------------*/
	/* setup 'inherit_mapp' */
	/*----------------------*/
	inherit_mapp = (Map *) ecalloc (sizeof (Map));
	inherit_mapp->attrs  = MA_INHERITED;
	inherit_mapp->nlinks = 1;		/* to 'protect' it */
}
/*}}}*/

/*{{{  create_default_proc_sp ()*/
create_default_proc_sp ()
/*---------------------*/
/*--------------------------------------------------------------*/
/*	Setup symtab entries for "$pN" (0 <= N <= MAX_RANK) for	*/
/* any default processor arrangements not specified by a command*/
/* line option 'pN', and attach them to the static (Symbol *)	*/
/* array 'default_proc_sp []'.					*/
/*--------------------------------------------------------------*/
{
	char    *def_proc_names = "$p0\0$p1\0$p2\0$p3\0$p4\0$p5\0$p6\0$p7";
			/* N.B. "$pN" means default shape, with 1 proc!
			   "$PN" means shape specified by command option. */
	int      dim,  rank;
	Array   *arp;
	Symbol  *sp;

	for (rank = 0;   rank <= MAX_RANK;   ++rank) {
		if (default_proc_sp [rank])
			continue;	/* shape specified by i/p option */

		sp = default_proc_sp [rank] =
		                  (Symbol *) ecalloc (sizeof (Symbol));

		sp->name = def_proc_names + 4*rank;	/* to point to "$pN",
							   where (N = rank) */
		sp->attrs = AT_PROCESSORS;
		if (rank) {
			arp = create_array ();
			for (dim = 0;  dim < rank;  ++dim)
				install_bounds (arp, NULL_Sym, NULL_Expr, int_lit_1);

			attach_array (sp->arp, arp);
		}
	}
}
/*}}}*/
/*}}}*/

/*{{{  << constructing / deleting 'Map' & 'Sect' structures >>*/
/**** In next 2, MA_INHERITED -> (MA_INHERITED | MA_ALIGNED_WITH_INHERIT) ? ***/

/*---------------------------------*/
/* constructing a 'Sect' structure */
/*---------------------------------*/
/*{{{  evaluate_subscrs_and_install_sect (src_ep, tgt_mapp) -- DONE VEC*/
evaluate_subscrs_and_install_sect (src_ep, tgt_mapp)
/*------------------------------------------------*/
	Expr  *src_ep;		/* ND_ARRAY_SECT /_WHOLE_ARRAY /_TEMP_ARRAY */
	Map   *tgt_mapp;	/* Target mapping for 'src_ep' -- used only
				     to determine the target mapping for
				     vector subscripts.  NULL if 'src_ep's
				     target mapping isn't decided yet. */
/*----------------------------------------------------------------
	Called by 'transform_and_write_assignment ()' and
'attain_mapping ()', to set-up an 'src_ep->sectp' structure for an
array, array element or array section 'src_ep'.  (If it's called by
'transform_and_write_assignment()',  'src_ep' is an assignment variable,
otherwise it's an expression primary.)

N.B.:  We assume that 'src_ep->mapp' is already installed!
----------------------------------------------------------------*/
{
	/*{{{  decls*/
	bool     contiguous;		/* i.e. 'sp' isn't itself a section! */
	int      array_rank,  dim,  sect_dim;
	Expr   **epp,  **l_p,  *l_norm,  *l_orig,  **n_p,	
	        *parent_l,  *parent_s,  *s_orig,  **s_p,
	       **sp_l_p,  **sp_n_p,  *subscr,  *subscrlist,  *u_orig;
	Map     *vec_mapp;
	Sect    *sectp,  *vec_sectp;
	Symbol  *sp;
	/*}}}*/

	if (src_ep->sectp)
		return;		/* already done! */

	sp = src_ep->val.sp;			   /* array's symtab entry */

	array_rank = sp->arp->attrs & RANK_MASK;   /* > 0 */

	/*--------------------------------------*/
	/* Create and install 'src_ep->sectp'	*/
	/*--------------------------------------*/
	sectp = (Sect *) ecalloc (sizeof (Sect));
	attach_sect (src_ep->sectp, sectp);

	sectp->attrs = array_rank;

	l_p = sectp->l = epp = (Expr **) emalloc (3 * array_rank * sizeof (Expr *));
	n_p = sectp->n = (epp += array_rank);
	s_p = sectp->s = (epp += array_rank);
	sp_l_p = sp->arp->l;
	sp_n_p = sp->arp->n;

	contiguous = ! (sp->mapp->attrs & MA_INHERITED);
	switch (src_ep->token) {
	case  ND_WHOLE_ARRAY :
	case  ND_TEMP_ARRAY :
		/*{{{  set all (l,n,s) = (0, n, 1)*/
		/*--------------------------------------------------------------*/
		/* A whole array reference.  Set all (l, n, s) = (0, n, 1)	*/
		/*--------------------------------------------------------------*/
		if (contiguous) {
			for (dim = 0;  dim < array_rank;
			                      ++dim,  ++l_p,  ++n_p,  ++s_p,  ++sp_n_p) {
		
				attach_expr (*l_p,  int_lit_0);
				attach_expr (*n_p, *sp_n_p);
				attach_expr (*s_p,  int_lit_1);
			}
		}
		else {
			for (dim = 0;  dim < array_rank;
			                      ++dim,  ++l_p,  ++n_p,  ++s_p,  ++sp_n_p) {
		
				attach_expr (*l_p,
				    map_primary (sp->name, 0, map_subscr_l (array_rank, dim)));
				attach_expr (*n_p, *sp_n_p);
				attach_expr (*s_p,
				    map_primary (sp->name, 0, map_subscr_s (array_rank, dim)));
			}
		}
		return;
		/*}}}*/

	case  ND_ARRAY_SECT :
/*************************************************
N.B. IT WOULD BE ADVANTAGEOUS TO DELETE (OR DETACH) SUBSCRS AS THEY'RE
PROCESSED.  NOT DOING SO MEANS THAT, IF SUBSCR IS REPLACED BY A TEMP
(E.G. A VEC SUBSCR) IT ISN'T DEALLOCATED A.S.A.P. AS IT HAS (nlinks > 0)!
**************************************************/
		/*{{{  set (l,n,s) for array element or section*/
		/*----------------------------------------------------------------
		An array element or section.  Ensure that all subscripts are mapped
		OK (i.e. that scalar subscripts and subscript triplets are globally
		replicated, and vector subscripts are comapped with the corresponding
		dimension of (tgt_mapp), and initialise the 'sectional' parameters
		sectp = (l[], n[], s[]).
		----------------------------------------------------------------*/
		for (subscrlist = src_ep->right,  dim = sect_dim = 0;
		     subscrlist  &&  (dim < array_rank);
		     subscrlist = subscrlist->right,  ++dim,  ++l_p,  ++n_p,  ++s_p,
		                                              ++sp_l_p,  ++sp_n_p) {
		
			subscr = subscrlist->left;	/* expr node for a subscript */
		
			if ( ! contiguous) {
				/*{{{  set 'parent_l/s'*/
				/*--------------------------------------------------------------*/
				/* N.B. These are garbage if unused, so 'protect' them below	*/
				/* and 'unprotect' them at the end (i.e. delete them if their	*/
				/* (nlinks == 0)).						*/
				/*--------------------------------------------------------------*/
				parent_l = map_primary (sp->name, 0, map_subscr_l (array_rank, dim));
				parent_s = map_primary (sp->name, 0, map_subscr_s (array_rank, dim));
				
				protect_expr (parent_l);
				protect_expr (parent_s);
				/*}}}*/
			}
		
			if (subscr->type & EX_SCALAR) {
				/*{{{  a scalar subscript (i)*/
				/*--------------------------------------------------------------*/
				/* A single subscript  (i)   (which, BTW, is always defined).	*/
				/* Then:							*/
				/*	l = { (i - sp_l)			-- contiguous	*/
				/*	    { parent_l + (i-sp_l)*parent_s	-- sectional	*/
				/*	n = -1							*/
				/*	s = 1		[This can be anything!]			*/
				/*--------------------------------------------------------------*/
				l_norm = expr (broadcast_scalar (subscr), ND_MINUS_OP, *sp_l_p);
				
				if ( ! contiguous)
					l_norm = expr (parent_l,
				                       ND_PLUS_OP,
				                       expr (l_norm, ND_TIMES_OP, parent_s));
				
				attach_expr (*l_p,  l_norm);
				attach_expr (*n_p,  int_lit_m1);
				attach_expr (*s_p,  int_lit_1);
				/*}}}*/
			}
			else {
				switch (subscr->token) {
				case  ND_SUBSCR_DOUBLET :
				case  ND_SUBSCR_TRIPLET :
					/*{{{  a subscript triplet ([l] : [u] [:s])*/
					/*----------------------------------------------------------------
					A subscript triplet ([l] : [u] [: s]).
						The expression tree pointed-to by 'subscr' has the general
					form:
					
					subscr  =  ND_SUBSCR_DOUBLET      or         ND_SUBSRC_TRIPLET
					              /          \                      /          \
					            [l]          [u]          ND_SUBSCR_DOUBLET     s
					                                         /          \
					                                       [l]          [u]
					
					where 'l' and/or 'u' (but not 's') may be NULL.
						A 'SUBSCR_TRIPLET' node is only created if [:s] is present,
					in which case 's' isn't NULL as that would be a syntax error!
					----------------------------------------------------------------*/
					
					/*{{{  set 's [dim]' & set 'subscr' to point to 'ND_DOUBLET' node (which exists!)*/
					/*--------------------------------------------------------------*/
					/* Set 's [dim]', and set 'subsrc' to point to the		*/
					/* 'ND_SUBSCR_DOUBLET' node (which always exists!).		*/
					/*--------------------------------------------------------------*/
					if (subscr->token == ND_SUBSCR_TRIPLET) {
						s_orig = broadcast_scalar (subscr->right);  /* s.  Has (nlinks > 0) */
						attach_expr (*s_p, (contiguous ?
						                        s_orig :
							                expr (s_orig, ND_TIMES_OP, parent_s)));
					
						subscr = subscr->left;		/* to point to the 'ND_SUBSCR_DOUBLET'
										   node (which always exists!) */
					}
					else {
						s_orig = int_lit_1;			/* save for later */
						attach_expr (*s_p, (contiguous ? int_lit_1 : parent_s));
					
						/* N.B. 'subscr' already points to the 'ND_SUBSCR_DOUBLET' node */
					}
					/*}}}*/
					
					/*--------------------------------------*/
					/* A subscript doublet ([l] : [u])	*/
					/*--------------------------------------*/
					if (subscr->left)  broadcast_scalar (subscr->left);		/* l */
					if (subscr->right) broadcast_scalar (subscr->right);		/* u */
					
					l_orig = subscr->left ?  subscr->left  : *sp_l_p;
					u_orig = subscr->right ? subscr->right : array_upper_bound (*sp_l_p, *sp_n_p);
					l_norm = subscr->left ?  expr (l_orig, ND_MINUS_OP, *sp_l_p) : int_lit_0;
					
					if ( ! contiguous)
						l_norm = expr (parent_l,
					                       ND_PLUS_OP,
					                       expr (l_norm, ND_TIMES_OP, parent_s));
					
					attach_expr (*l_p,  l_norm);
					attach_expr (*n_p,  sect_extent_func (l_orig, u_orig, s_orig));
					
					/*--------------------------------------------------------------*/
					/* N.B.:							*/
					/* 'l_orig' has (nlinks > 0), so it's protected.		*/
					/* 'u_orig' may have (nlinks == 0), but is only used once, so ok*/
					/* 'l_norm' may have (nlinks == 0), but is only used once, so ok*/
					/*--------------------------------------------------------------*/
					break;
					/*}}}*/
		
				default :
					/*{{{  a vector subscript*/
					/*--------------------------------------------------------------*/
					/* A vector subscript  (V).  Then:				*/
					/*	l = { (V - sp_l)			-- contiguous	*/
					/*	    { parent_l + (V-sp_l)*parent_s	-- sectional	*/
					/*	n = -2							*/
					/*	s = 1		[This can be anything!]			*/
					/*--------------------------------------------------------------*/
					if (tgt_mapp)
						extract_dim_from_map (&vec_mapp, &vec_sectp, tgt_mapp, sect_dim);
					else
						choose_map_for_vec_subscr (&vec_mapp, &vec_sectp, src_ep, subscr, dim);
					
					l_norm = expr (subscr, ND_MINUS_OP, *sp_l_p);
					
					if ( ! contiguous)
						l_norm = expr (parent_l,
					                       ND_PLUS_OP,
					                       expr (l_norm, ND_TIMES_OP, parent_s));
					
					attach_expr (*l_p,  attain_mapping (l_norm, vec_mapp, vec_sectp, RESULT_VAR));
					attach_expr (*n_p,  int_lit_m2);
					attach_expr (*s_p,  int_lit_1);
					
					break;
					/*}}}*/
				}
				++sect_dim;
			}
				
			if ( ! contiguous) {
				/*{{{  delete 'parent_l/s' if they're unused*/
				unprotect_expr (parent_l);
				unprotect_expr (parent_s);
				/*}}}*/
			}
		}
		
		if ((dim < array_rank)  ||  subscrlist) {
			errmsg ("reference to array %q has too %s dimensions",
			                   sp->name, (subscrlist ? "many" : "few"));
			/*------------------------------------------------------*/
			/* For safety, supply arbitrary values for any missing	*/
			/* (l,n,s);  set them to (0, -1, 1).			*/
			/*------------------------------------------------------*/
			for (  ;  dim < array_rank;    ++dim,  ++l_p,  ++n_p,  ++s_p) {
				attach_expr (*l_p,  int_lit_0);
				attach_expr (*n_p,  int_lit_m1);
				attach_expr (*s_p,  int_lit_1);
			}
		}
		
		return;
		/*}}}*/
	}
}
/*}}}*/
/*{{{  Sect  *whole_object_sect (mapp, sp)*/
Sect  *whole_object_sect (mapp, sp)
/*-------------------------------*/
	Map     *mapp;		/* != NULL */
	Symbol  *sp;		/* NULL for temp arrays (but that's ok as only
				   used for 'sectional' dummy arg arrays) */
/*--------------------------------------------------------------*/
/*	If 'mapp' is the mapping for an *array*, create and	*/
/* return a 'Sect' structure for a reference to a *whole* array	*/
/* with this mapping;  otherwise return 'NULL_Sect'.		*/
/*--------------------------------------------------------------*/
{
	/*{{{  decls*/
	bool    contiguous;
	int     array_rank,  dim;
	Expr  **an_p,  **epp,  **l_p,  **n_p,  **s_p,  **sp_n_p;
	Sect   *sectp;
	/*}}}*/

	if (! mapp  ||
	    ! (array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT)) )

		return  NULL_Sect;

	/*-------------------------------*/
	/* Create and initialise 'sectp' */
	/*-------------------------------*/
	sectp = (Sect *) ecalloc (sizeof (Sect));
	sectp->attrs = array_rank;
	l_p = sectp->l = epp = (Expr **) emalloc (3 * array_rank * sizeof (Expr *));
	n_p = sectp->n = (epp += array_rank);
	s_p = sectp->s = (epp += array_rank);

	contiguous = ! (mapp->attrs & MA_INHERITED);
	if (contiguous) {
		/*------------------------------------------------*/
		/* A contiguous array; set all (l,n,s) = (0,an,1) */
		/*------------------------------------------------*/
		an_p = mapp->an;
		for (dim = 0;   dim < array_rank;
		              ++dim,  ++l_p,  ++n_p,  ++s_p,  ++an_p) {

			attach_expr (*l_p,  int_lit_0);
			attach_expr (*n_p, *an_p);
			attach_expr (*s_p,  int_lit_1);
		}
	}
	else {	/*--------------------------------------*/
		/* Array 'sp' is itself a section.	*/
		/*--------------------------------------*/
		sp_n_p = sp->arp->n;
		for (dim = 0;   dim < array_rank;
		              ++dim,  ++l_p,  ++n_p,  ++s_p,  ++sp_n_p) {

			attach_expr (*l_p,
			    map_primary (sp->name, 0, map_subscr_l (array_rank, dim)));
			attach_expr (*n_p, *sp_n_p);
			attach_expr (*s_p,
			    map_primary (sp->name, 0, map_subscr_s (array_rank, dim)));
		}
	}
	return  sectp;
}

/*}}}*/
/*{{{  restore_whole_array_sect (temp_ep)*/
restore_whole_array_sect (temp_ep)
/*------------------------------*/
	Expr    *temp_ep;	/* ...a 'ND_TEMP_ARRAY' node.  It has
				   'mapp' and 'sectp' defined. */
/*----------------------------------------------------------------
	Called by 'write_assignment_stmt ()' after it has
written-out an assignment to the temporary array 'temp_ep',
to restore its 'Sect' record 'temp_ep->sectp' to its original
state (viz. that created by 'whole_object_sect ()') before being
changed (i.e. 'localised') by 'localise_array_expr ()'.
	We choose to restore the original 'temp_ep->sectp' record,
rather replacing it by a new one, in case it's used elsewhere
(e.g. as a 'tgt_sectp').

N.B.  We assume 'temp_ep' occupies *contiguous* storage (i.e. it
isn't a section of a larger array).
----------------------------------------------------------------*/
{
	int     array_rank,  dim;
	Expr  **anp,  **lp,  **np,  **sp;
	Map    *mapp  = temp_ep->mapp;		/* != NULL */
	Sect   *sectp = temp_ep->sectp;		/* != NULL */

	array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);

	/*---------------------------------------*/
	/* Set all sectp->(l, n, s) = (0, an, 1) */
	/*---------------------------------------*/
	lp = sectp->l;
	np = sectp->n;
	sp = sectp->s;
	anp = mapp->an;		/* 'protected' */

	clear_flags (sectp->attrs, SE_LOCALISED_SECT);

	for (dim = 0;   dim < array_rank;
	              ++dim,  ++lp,  ++np,  ++sp,  ++anp) {
	
		delete_expr (*lp, DECR_NLINKS);
		delete_expr (*np, DECR_NLINKS);
		delete_expr (*sp, DECR_NLINKS);

		attach_expr (*lp,  int_lit_0);
		attach_expr (*np, *anp);
		attach_expr (*sp,  int_lit_1);
	}
}
/*}}}*/

/*-----------------------------------------------------------*/
/* constructing 'Map' (and 'Sect' if appropriate) structures */
/*-----------------------------------------------------------*/
/*{{{  compose_map_and_sect (new_mapp_p, new_sectp_p, sect_ep) -- OK*/
compose_map_and_sect (new_mapp_p, new_sectp_p, sect_ep)
/*---------------------------------------------------*/
	Map   **new_mapp_p;	/* result parameter */
	Sect  **new_sectp_p;	/*   "       "      */
	Expr   *sect_ep;	/* ND_WHOLE_ARRAY, ND_ARRAY_SECT, or
				   ND_SIMPLE_SCALAR.  'sect_ep->sectp'
				   & 'mapp' may be NULL. */
/*{{{  < DESCRIPTION >*/
/*----------------------------------------------------------------
	Combine 'sect_ep->mapp' and 'sect_ep->sectp' (if defined)
to produce a 'rationalised' 'new_mapp',  i.e. such that the *whole*
object described by 'new_mapp' is equivalent in rank, shape and
mapping to the section described by 'ep->sectp' and 'ep->mapp'
together.  On return, the parameters 'new_mapp_p' and 'new_sectp_p'
are set to point to the newly created structures 'new_mapp' and
'new_sectp' respectively (the latter set to select the *whole* array
described by 'new_mapp').
	The composition is performed as follows:

sectp->(l,n,s)
--------------
(i, -1, s)  -- i.e. a single element selected

	If it's a collapsed dim, (d == -1), omit it.  Otherwise
	convert it to an embedded dim.

(v, -2, ?)  -- i.e. an irregular section with vector subscript 'v'

	Let 'N' be the length of vector 'v'.  Then:
		(al',an',as',tn',b',fmt',d') = (0,N,1,N, b",fmt",d),
	where  fmt" = fmt,  and b" follows from fmt".

(l, n, s)   -- (n != -1 or -2);  i.e. a regular section of 'n' elements

	If it's a collapsed dim, (d == -1):
		(al',an',as',tn',b',fmt',d') = (0,n,1,n,n,COLLAPSED,-1)
	Otherwise combine (l,n,s) with (al,an,as) to produce
	(al',an',as').  Leave (tn, b, fmt, d) as before.


N.B.  If 'ep' is array-valued, this always creates and returns
*new*  'Map' and 'Sect' structures.  However, if it's scalar,
this just returns the existing 'ep->mapp' (if there is one).
----------------------------------------------------------------*/
/*}}}*/
{
	/*{{{  decls*/
	int     array_rank,  dim,  embed_rank,  new_array_rank,  new_embed_rank;
	int32   attrs,  new_attrs;
	short  *dp;
	Expr  **np;
	Map    *mapp,  *new_mapp;
	Sect   *sectp;
	/*}}}*/

	mapp  = sect_ep->mapp;
	sectp = sect_ep->sectp;
	if ( ! (mapp  &&  sectp)) {	/* simple scalar (replic or embedded)*/
		/*{{{  return (*new_mapp_p = mapp;  *new_sectp_p = sectp)*/
		/*--------------------------------------------------------------*/
		/* A simple scalar, either:					*/
		/* -- globally replicated  (mapp == sectp == NULL), or		*/
		/* -- embedded             (mapp != NULL;  sectp == NULL).	*/
		/*	In either case return (*new_mapp_p = mapp;		*/
		/* *new_sectp_p = sectp).					*/
		/*--------------------------------------------------------------*/
		*new_mapp_p  = mapp;
		*new_sectp_p = sectp;
		return;
		/*}}}*/
	}
	/*-----------------------------------------*/
	/* An array-valued object or array element */
	/*-----------------------------------------*/
	/*{{{  find 'attrs', 'array_rank' & 'embed_rank'*/
	attrs = mapp->attrs;
	array_rank  = ((attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	embed_rank  = ((attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	/*}}}*/
	/*{{{  find 'new_array_rank', 'new_embed_rank' & 'new_attrs'*/
	/*----------------------------------------------------------------
		Find 'new_array_rank' and 'new_embed_rank', allowing for
	the fact that array dimensions with a single element selected are:
	-- converted to embedded dims if they're distributed;
	-- omitted if they're collapsed;
	
		Therefore we compute 'new_array_rank' and 'new_embed_rank'
	as follows:
	-- initialise them to 'array_rank' and 'embed_rank'
	-- for dims with (n == -1), --new_array_rank
	-- if (n == -1 && d != -1), ++new_embed_rank.
	----------------------------------------------------------------*/
	new_array_rank = array_rank;
	new_embed_rank = embed_rank;
	new_attrs = attrs;
	
	np = sectp->n;
	dp = mapp->d;
	
	for (dim = 0;    dim < array_rank;    ++dim,  ++np,  ++dp) {
		if (*np == int_lit_m1) {
	
			--new_array_rank;
	
			new_attrs -= (1 << MA_ARRAY_RANK_SHIFT);
	
			if (*dp != -1) {
				/*--------------------------------------*/
				/* convert this distributed array dim	*/
				/* to an embedded dim.			*/
				/*--------------------------------------*/
				++new_embed_rank;
	
				new_attrs += (1 << MA_EMBED_RANK_SHIFT);
			}
			/*---------------------------------------------*/
			/* ...else collapsed, so forget it altogether! */
			/*---------------------------------------------*/
		}
	}
	/*}}}*/

	if (! (new_array_rank || new_embed_rank) && (new_attrs & MA_GLOBALLY_REPLIC)) {
		/*{{{  a global scalar;  return (*new_mapp_p = *new_sectp_p = NULL)*/
		/*--------------------------------------------------------------*/
		/* The result is a globally replicated simple scalar		*/
		/* (presumably arising from an array element of a globally	*/
		/* replicated array).						*/
		/*	Return (*new_mapp_p = *new_sectp_p = NULL).		*/
		/*--------------------------------------------------------------*/
		*new_mapp_p  = NULL_Map;
		*new_sectp_p = NULL_Sect;
		return;
		/*}}}*/
	}

	/*{{{  create & initialise 'new_mapp'*/
	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	new_mapp = (Map *) ecalloc (sizeof (Map));
	
	new_mapp->attrs  = new_attrs;
	new_mapp->nlinks = 0;		/* as not attached to anything yet! */
	
	if (new_array_rank) {
		/*{{{  install (al, an, as, b, tn, fmt, d)  -- ok*/
		/*------------------------------------------------------*/
		/* Install  (al[], an[], as[], tn[], b[], fmt[], d[])	*/
		/*------------------------------------------------------*/
		Expr  **epp,  **lp,  **sp,  **alp,  **anp,  **asp,  **tnp,  **bp,
		      **new_alp,  **new_anp,  **new_asp,  **new_tnp,  **new_bp;
		short  *fmtp,  *new_fmtp,  *new_dp;
		
		
		new_alp = new_mapp->al =  epp = (Expr **) emalloc (5 * new_array_rank * sizeof (Expr *));
		new_anp = new_mapp->an = (epp += new_array_rank);
		new_asp = new_mapp->as = (epp += new_array_rank);
		new_tnp = new_mapp->tn = (epp += new_array_rank);
		new_bp  = new_mapp->b  = (epp += new_array_rank);
		
		new_fmtp = new_mapp->fmt = (short *) emalloc (2 * new_array_rank * sizeof (short));
		new_dp   = new_mapp->d   = (new_fmtp + new_array_rank);
		
		lp  = sectp->l;		/* 'protected' */
		np  = sectp->n;		/*      "      */
		sp  = sectp->s;		/*      "      */
		
		alp = mapp->al;		/*      "      */
		anp = mapp->an;		/*      "      */
		asp = mapp->as;		/*      "      */
		tnp = mapp->tn;		/*      "      */
		bp  = mapp->b;		/*      "      */
		fmtp = mapp->fmt;	/*      "      */
		dp   = mapp->d;		/*      "      */
		
		for (dim = 0;   dim < array_rank;
		   ++dim, ++lp, ++np, ++sp, ++alp, ++anp, ++asp, ++tnp, ++bp, ++fmtp, ++dp) {
		
			if (*np == int_lit_m1)
				continue;	/* single element; not counted as an
						   'array dim' in 'new_mapp' */
			if (*np == int_lit_m2) {
				/*{{{  an irregular section*/
				/*------------------------------------------------------*/
				/* An irregular section, with vector subscript '*lp'.	*/
				/*------------------------------------------------------*/
				Expr  *n_vec;
				
				/*------------------------------------------------------*/
				/* Find 'n_vec', the length of vector subscript '*lp'	*/
				/*------------------------------------------------------*/
				n_vec = vec_len (find_array_term (*lp));
							/* N.B.  The array term passed to 'vec_len ()' has
							   'mapp' (and if necessary 'sectp') installed */
				
				/*----------------------------------------------------------------
				(al,an,as, tn, b, fmt,d)  =  (0, n_vec, 1, n_vec, b', fmt', d),
				where:
					fmt'  = fmt (src_ep [src_dim]),  except that CYCLIC -> BLOCK
					b'    = { n_vec			-- if (fmt' == COLLAPSED)
					        { CEILING (n_vec / np)	-- if (fmt' == BLOCK)
				----------------------------------------------------------------*/
				attach_expr (*new_alp,  int_lit_0);
				attach_expr (*new_anp,  n_vec);
				attach_expr (*new_asp,  int_lit_1);
				attach_expr (*new_tnp,  n_vec);
				
				*new_dp = *dp;	/*** WHAT IF == -2 ?? ******/
				
				switch (*fmtp) {
				case  DFMT_COLLAPSED :
					*new_fmtp = DFMT_COLLAPSED;
					attach_expr (*new_bp,  n_vec);
					break;
				
				case  DFMT_BLOCK :
				case  DFMT_CYCLIC :
				{
					/*******************************************/
					/****  WHAT IF INHERITED PROCS ???  ********/
					/*******************************************/
					int    procdim = proc_dim (mapp->fmt, dim);
					Expr  *n_procs = mapp->proc_sp->arp->n [procdim];
				
					*new_fmtp = DFMT_BLOCK;
					attach_expr (*new_bp, ceiling_div (n_vec, n_procs));
					break;
				}
				case  DFMT_INHERIT :
					/***********************************/
					/****  WHAT TO DO HERE ???  ********/
					/***********************************/
					break;
				}
				
				
				/*}}}*/
			}
			else {
				/*{{{  a regular section*/
				/*----------------------*/
				/* A regular section.	*/
				/*----------------------*/
				/*------------------------------------------------------*/
				/* new_(an, fmt, d) == original ones in all cases...	*/
				/*------------------------------------------------------*/
				attach_expr (*new_anp, *np);	/* new_an = 'n' in all cases */
				*new_fmtp = *fmtp;
				*new_dp   = *dp;
				
				/*--------------------------------------*/
				/* Now evaluate  new_(al, as, tn, b)	*/
				/*--------------------------------------*/
				if (*dp == -1) {
					/*{{{  collapsed:  new_(al, as, tn, b) = (0, 1, n, n)*/
					/*----------------------------------------------*/
					/* A collapsed dimension.  Set its params to:	*/
					/*	new_(al, as, tn, b) = (0, 1, n, n)	*/
					/*----------------------------------------------*/
					attach_expr (*new_alp, int_lit_0);
					attach_expr (*new_asp, int_lit_1);
					attach_expr (*new_tnp, *np);
					attach_expr (*new_bp,  *np);
					/*}}}*/
				}
				else {
					/*{{{  not collapsed:  combine sections*/
					/*------------------------------------------------------*/
					/* A section of a distributed dimension.  Combine the	*/
					/* parameters as follows:				*/
					/*	new_al  = al + (l * as)				*/
					/*	new_as  = (s * as)				*/
					/*	new_tn  = tn					*/
					/*	new_b   = b					*/
					
					/****** COULD SIMPLIFY FURTHER BY DIVIDING BY hcf (b, |as)) ********/
					/*------------------------------------------------------*/
					attach_expr (*new_alp, expr (*alp, ND_PLUS_OP, expr (*lp, ND_TIMES_OP, *asp)));
					attach_expr (*new_asp, expr (*sp, ND_TIMES_OP, *asp));
					attach_expr (*new_tnp, *tnp);
					attach_expr (*new_bp,  *bp);
					/*}}}*/
				}
				/*}}}*/
			}
			++new_alp;   ++new_anp;   ++new_asp;   ++new_tnp;   ++new_bp;
			++new_fmtp;  ++new_dp;
		}
		/*}}}*/
	}
	
	if (new_embed_rank) {
		/*{{{  install (ep, ed) -- ok*/
		Expr  **epp,  **new_epp;
		short  *edp,  *new_edp;
		
		new_epp = new_mapp->ep = (Expr **) emalloc (new_embed_rank * sizeof (Expr *));
		new_edp = new_mapp->ed = (short *) emalloc (new_embed_rank * sizeof (short));
		
		/*--------------------------------------------------------------*/
		/* Convert all distributed array dimensions with a single	*/
		/* element selected into embedded dimensions.			*/
		/*--------------------------------------------------------------*/
		np = sectp->n;		/* 'protected' */
		dp = mapp->d;		/*      "      */
		
		for (dim = 0;    dim < array_rank;    ++dim,  ++np,  ++dp) {
		
			if ((*np != int_lit_m1) || (*dp == -1))
				continue;	/* a regular section or a single element
						   in a collapsed dimension.	*/
		
			/*------------------------------------------------------*/
			/* ...else a single element in a distributed dimension.	*/
			/* Convert this to an embedded dimension.		*/
			/*------------------------------------------------------*/
			proc_and_subscript (new_epp, NULL_Expr_p, mapp, sectp->l [dim], dim);
			*new_edp = *dp;
		
			++new_epp;  ++new_edp;
		}
		
		/*--------------------------------------------------------------*/
		/* Now copy the original embedded dimensions to 'new_mapp'.	*/
		/*--------------------------------------------------------------*/
		epp = mapp->ep;
		edp = mapp->ed;
		
		for (dim = 0;   dim < embed_rank;   ++dim, ++epp, ++edp, ++new_epp, ++new_edp) {
			attach_expr (*new_epp,  *epp);
			*new_edp = *edp;
		}
		/*}}}*/
	}
	
	new_mapp->proc_sp = mapp->proc_sp;  	/* NULL for 'inherit procs'*/
	/*}}}*/

	*new_mapp_p  = new_mapp;
	*new_sectp_p = whole_object_sect (new_mapp, NULL_Sym);
	return;
}
/*}}}*/

/*{{{  extract_dim_from_map (new_mapp_p, new_sectp_p, mapp, fr_dim)--OK*/
extract_dim_from_map (new_mapp_p, new_sectp_p, mapp, fr_dim)
/*--------------------------------------------------------*/
	Map   **new_mapp_p;		/* result parameter */
	Sect  **new_sectp_p;		/*   "       "      */
	Map    *mapp;			/* != NULL */
	int     fr_dim;
/*----------------------------------------------------------------
	Called by 'evaluate_subscrs_and_install_sect ()' to
evaluate the mapping required of a vector subscript.  It creates
and returns new mapping structures (new_mapp, new_sectp) for a
1-d array with the same mapping as dimension 'fr_dim' of (mapp).

N.B. 'mapp' shouldn't be NULL and should have dimension 'fr_dim'
defined.  However, we check these properties to allow for errors
in the input code. 
----------------------------------------------------------------*/
/*** THIS DOESN'T TEST WHETHER mapp HAS dim 'fr_dim' *******/
{
	/*{{{  decls*/
	int32   attrs;
	int     array_rank,  dim,  embed_rank;
	short  *dp;
	Expr  **epp;
	Map    *new_mapp;
	/*}}}*/

	/*{{{  find 'array_rank' & 'embed_rank'*/
	attrs = mapp->attrs;
	
	array_rank  = ((attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	embed_rank  = ((attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	
	/*}}}*/
	
	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	new_mapp  = (Map *) ecalloc (sizeof (Map));

	new_mapp->nlinks = 0;		/* ...not attached to anything yet! */

	/*{{{  copy (al, an, as, tn, b, fmt, d) from dimension 'fr_dim' of 'mapp'*/
	/*--------------------------------------------------------------*/
	/* Copy (al, an, as, tn, b, fmt, d) from dimension 'fr_dim' of	*/
	/* 'mapp'.							*/
	/*--------------------------------------------------------------*/
	new_mapp->al =   epp = (Expr **) emalloc (5 * sizeof (Expr *));
	new_mapp->an = ++epp;
	new_mapp->as = ++epp;
	new_mapp->tn = ++epp;
	new_mapp->b  = ++epp;
	
	new_mapp->fmt =   dp = (short *) emalloc (2 * sizeof (short));
	new_mapp->d   = ++dp;
	
	attach_expr (new_mapp->al [0],  mapp->al [fr_dim]);
	attach_expr (new_mapp->an [0],  mapp->an [fr_dim]);
	attach_expr (new_mapp->as [0],  mapp->as [fr_dim]);
	attach_expr (new_mapp->tn [0],  mapp->tn [fr_dim]);
	attach_expr (new_mapp->b  [0],  mapp->b  [fr_dim]);
	
	new_mapp->fmt [0] = mapp->fmt [fr_dim];
	new_mapp->d [0]   = mapp->d [fr_dim];
	/*}}}*/

	if (embed_rank) {
		/*{{{  copy (ep, ed) from 'mapp' to 'new_mapp'*/
		/*------------------------------------------------------*/
		/* Copy the original embedded dimensions to 'new_mapp'.	*/
		/*------------------------------------------------------*/
		short  *edp,   *new_edp;
		Expr  **new_epp;
		
		new_epp = new_mapp->ep = (Expr **) emalloc (embed_rank * sizeof (Expr *));
		new_edp = new_mapp->ed = (short *) emalloc (embed_rank * sizeof (short));
		
		epp = mapp->ep;
		edp = mapp->ed;
		
		for (dim = embed_rank;    dim;
		                             --dim,  ++epp,  ++edp,  ++new_epp,  ++new_edp) {
			attach_expr (*new_epp,  *epp);
			*new_edp = *edp;
		}
		/*}}}*/
	}
	new_mapp->proc_sp = mapp->proc_sp;

	/*{{{  set 'new_mapp->attrs'*/
	new_mapp->attrs =  MA_INTEGER |
	                  (1 << MA_ARRAY_RANK_SHIFT) |
	                  (embed_rank << MA_EMBED_RANK_SHIFT) |
	                  (attrs & MA_INHERIT_MASK);
	
	if ( (new_mapp->fmt [0] != DFMT_COLLAPSED)  ||  embed_rank  ||
	     (attrs & (MA_INHERITED | MA_ALIGNED_WITH_INHERIT | MA_INHERIT_FMT)) )
	
		new_mapp->attrs |= MA_DISTRIB;
	
	else if (new_mapp->proc_sp  ||  (attrs & MA_INHERIT_PROCS))
	
		new_mapp->attrs |= MA_REPLIC_ON_PROCS;
	
	else
		new_mapp->attrs |= MA_GLOBALLY_REPLIC;
	/*}}}*/

	*new_mapp_p  = new_mapp;
	*new_sectp_p = whole_object_sect (new_mapp, NULL_Sym);
	return;
}
/*}}}*/
/*{{{  omit_dim_from_map (new_mapp_p, new_sectp_p, mapp, dim_ep) -- check*/
omit_dim_from_map (new_mapp_p, new_sectp_p, mapp, dim_ep)
/*-----------------------------------------------------*/
	Map   **new_mapp_p;		/* result parameter */
	Sect  **new_sectp_p;		/*   "       "      */
	Map    *mapp;			/* != NULL */
	Expr   *dim_ep;			/* N.B. uses Fortran convention, i.e.
					   (dim == 1) for lowest dim ! */
/*----------------------------------------------------------------
	Called by 'transform_reduction_func () / _SHIFT () / _loc_func ()'
to create and return new mapping structures (new_mapp, new_sectp)
that have the same mapping as 'mapp' but with dimension '*dim_ep'
omitted  (or more precisely, if dimension '*dim_ep' of 'tgt_mapp'
was distributed over a processor array dimension, then 'new_mapp'
is replicated over that processor array dimension).

N.B. 'mapp' shouldn't be NULL and should have dimension '*dim_ep'
defined.  However, we check these properties to allow for errors
in the input code. 
----------------------------------------------------------------*/
/*** THIS DOESN'T TEST WHETHER mapp HAS DIM (omit_dim = *dim_ep) *******/
{
	/*{{{  decls*/
	bool    distr_dims;
	int32   attrs;
	int     array_rank,  dim,  embed_rank,  new_array_rank,  omit_dim;
	Map    *new_mapp;
	/*}}}*/

	if (dim_ep->type & EX_INT_LIT_CONST)
		omit_dim = int_val (dim_ep) - 1;   /* ...-1 so starts from 0!*/
	else {
		not_in_subset ("non-constant \'DIM\' argument of an intrinsic function");
		omit_dim = 0;		/* as a fix! */
	}

	/*{{{  find 'array_rank', 'embed_rank' & 'new_array_rank'*/
	attrs = mapp->attrs;
	
	array_rank  = ((attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	embed_rank  = ((attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	
	new_array_rank = array_rank - 1;
	/*}}}*/

	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	new_mapp = (Map *) ecalloc (sizeof (Map));

	new_mapp->nlinks = 0;		/* ...not attached to anything yet! */

	distr_dims = FALSE;
	if (new_array_rank) {
		/*{{{  copy (al,an,as,tn,b,fmt,d) from 'mapp', excluding dim 'omit_dim'*/
		/*--------------------------------------------------------------*/
		/* Copy (al[], an[], as[], tn[], b[], fmt[], d[]) from 'mapp',	*/
		/* excluding dimension 'omit_dim'.				*/
		/*--------------------------------------------------------------*/
		Expr  **epp,  **alp,  **anp,  **asp,  **tnp,  **bp,
		      **new_alp,  **new_anp,  **new_asp,  **new_tnp,  **new_bp;
		short  *dp,  *fmtp,  *new_dp,  *new_fmtp;
		
		
		new_alp = new_mapp->al =  epp = (Expr **) emalloc (5 * new_array_rank * sizeof (Expr *));
		new_anp = new_mapp->an = (epp += new_array_rank);
		new_asp = new_mapp->as = (epp += new_array_rank);
		new_tnp = new_mapp->tn = (epp += new_array_rank);
		new_bp  = new_mapp->b  = (epp += new_array_rank);
		
		new_fmtp = new_mapp->fmt = (short *) emalloc (2 * new_array_rank * sizeof (short));
		new_dp   = new_mapp->d   = (new_fmtp + new_array_rank);
		
		alp = mapp->al;
		anp = mapp->an;
		asp = mapp->as;
		tnp = mapp->tn;
		bp  = mapp->b;
		
		fmtp = mapp->fmt;
		dp   = mapp->d;
		
		for (dim = 0;   dim < array_rank;
		   ++dim, ++alp, ++anp, ++asp, ++tnp, ++bp, ++fmtp, ++dp) {
		
			if (dim == omit_dim)
				continue;	/* don't copy this dim! */
		
			attach_expr (*new_alp, *alp);
			attach_expr (*new_anp, *anp);
			attach_expr (*new_asp, *asp);
			attach_expr (*new_tnp, *tnp);
			attach_expr (*new_bp,  *bp);
		
			*new_fmtp = *fmtp;
			*new_dp   = *dp;
		
			if (*new_fmtp != DFMT_COLLAPSED)
				distr_dims = TRUE;
		
			++new_alp;   ++new_anp;   ++new_asp;   ++new_tnp;   ++new_bp;
			++new_fmtp;  ++new_dp;
		}
		/*}}}*/
	}
	if (embed_rank) {
		/*{{{  copy (ep, ed) from 'mapp' to 'new_mapp'*/
		/*------------------------------------------------------*/
		/* Copy the original embedded dimensions to 'new_mapp'.	*/
		/*------------------------------------------------------*/
		short  *edp,   *new_edp;
		Expr  **epp,  **new_epp;
		
		new_epp = new_mapp->ep = (Expr **) emalloc (embed_rank * sizeof (Expr *));
		new_edp = new_mapp->ed = (short *) emalloc (embed_rank * sizeof (short));
		
		epp = mapp->ep;
		edp = mapp->ed;
		
		for (dim = embed_rank;   dim;   --dim, ++epp, ++edp, ++new_epp, ++new_edp) {
			attach_expr (*new_epp,  *epp);
			*new_edp = *edp;
		}
		/*}}}*/
	}
	new_mapp->proc_sp = mapp->proc_sp;

	/*{{{  set 'new_mapp->attrs'*/
	new_mapp->attrs = (attrs & (MA_TYPE | MA_INHERIT_MASK))  |
	                  (new_array_rank << MA_ARRAY_RANK_SHIFT)  |
	                  (embed_rank << MA_EMBED_RANK_SHIFT);
	
	if (  distr_dims  ||  embed_rank  ||
	     (attrs & (MA_INHERITED | MA_ALIGNED_WITH_INHERIT | MA_INHERIT_FMT)) )
	
		new_mapp->attrs |= MA_DISTRIB;
	
	else if (new_mapp->proc_sp  ||  (attrs & MA_INHERIT_PROCS))
	
		new_mapp->attrs |= MA_REPLIC_ON_PROCS;
	
	else
		new_mapp->attrs |= MA_GLOBALLY_REPLIC;
	/*}}}*/

	*new_mapp_p  = new_mapp;
	*new_sectp_p = whole_object_sect (new_mapp, NULL_Sym);
	return;
}
/*}}}*/
/*{{{  global_array_map (mapp_p, sectp_p, type, rank, shape) -- check*/
global_array_map (mapp_p, sectp_p, type, rank, shape)
/*-------------------------------------------------*/
	Map   **mapp_p;		/* result parameter */
	Sect  **sectp_p;	/*   "       "      */
	int32   type;		/* data type of result array */
	int     rank;		/* >0 */
	Expr  **shape;		/* shape vector 'shape[]', of length 'rank'*/
/*----------------------------------------------------------------
	Called by 'transform_loc_func ()' to create and return
new mapping structures (mapp, sectp) that describe a
globally replicated array with rank 'rank' and shape 'shape[]'.
----------------------------------------------------------------*/
{
	int     dim;
	short  *dp,  *fmtp;
	Expr  **epp,  **alp,  **anp,  **asp,  **np,  **tnp,  **bp;
	Map    *mapp;

	/*--------------------------------------*/
	/* Create and initialise 'mapp'	*/
	/*--------------------------------------*/
	mapp = (Map *) ecalloc (sizeof (Map));

	mapp->attrs = (type  |  (rank << MA_ARRAY_RANK_SHIFT)  |
	               MA_GLOBALLY_REPLIC);

	/*{{{  install (al, an, as, tn, b, fmt, d)*/
	/*--------------------------------------------------------------*/
	/* Install (al[], an[], as[], tn[], b[], fmt[], d[]).		*/
	/* For each dimension:						*/
	/*	(al,an,as, tn,b, fmt,d) = (0,n,1, n,n, COLLAPSED,-1)	*/
	/*--------------------------------------------------------------*/
	alp = mapp->al =  epp = (Expr **) emalloc (5 * rank * sizeof (Expr *));
	anp = mapp->an = (epp += rank);
	asp = mapp->as = (epp += rank);
	tnp = mapp->tn = (epp += rank);
	bp  = mapp->b  = (epp += rank);
	
	fmtp = mapp->fmt = (short *) emalloc (2 * rank * sizeof (short));
	dp   = mapp->d   = (fmtp + rank);
	
	np = shape;
	
	for (dim = 0;    dim < rank;    ++dim,  ++np) {
	
		attach_expr (*alp++,  int_lit_0);
		attach_expr (*anp++, *np);
		attach_expr (*asp++,  int_lit_1);
		attach_expr (*tnp++, *np);
		attach_expr (*bp++,  *np);
	
		*fmtp++ = DFMT_COLLAPSED;	/* collapsed */
		*dp++   = -1;			/*     "     */
	}
	/*}}}*/

	*mapp_p  = mapp;
	*sectp_p = whole_object_sect (mapp, NULL_Sym);
	return;
}
/*}}}*/


/****************************************************************
In next 2, still have to allow for 'find_array_term()' returing
'ND_FUNC' for an array-valued external or transformational function!
	Need something like:
		evaluate_function_and_install_sect (func_ep) !
For external funcs, lift code from 'attain_mapping()' but don't remap
result (since no tgt!).  For transformational ??
(IN FACT -- DO THIS ANYWAY -- INSERT CALL TO THAT IN 'attain_mapping()'??
*******************************************************************/
/*{{{  choose_map_for_vec_subscr (vec_mapp_p, vec_sectp_p, src_ep, vec_ep, src_dim)--OK*/
choose_map_for_vec_subscr (vec_mapp_p, vec_sectp_p, src_ep, vec_ep, src_dim)
/*------------------------------------------------------------------------*/
	Map   **vec_mapp_p;	/* result parameter */
	Sect  **vec_sectp_p;	/*   "       "      */
	Expr   *src_ep;		/* 'ND_ARRAY_SECT' or 'ND_WHOLE_ARRAY' */
	Expr   *vec_ep;		/* vec subscript in dim 'src_dim' of 'src_ep'*/
	int     src_dim;
/*----------------------------------------------------------------
	Called by 'evaluate_subscrs_and_install_sect ()', in the
case where there is no target mapping for the array section
concerned, to create and return a mapping (vec_mapp, vec_sectp)
suitable for vector subscript 'vec_ep', which is the subscript in
dimension 'src_dim' of 'src_ep'.  The chosen mapping is that of a
1-d array of (n = vec_len (vec_ep)) elements distributed as uniformly
as possible over the same processor array dimension as dimension
'src_dim' of 'src_ep'.

N.B. 'src_mapp' shouldn't be NULL and should have dimension 'src_dim'
defined.  However, we check these properties to allow for errors
in the input code. 
----------------------------------------------------------------*/
/*****************************
BUG
---
THIS DOESN'T ALLOW FOR 'find_array_term()' RETURNING A 'ND_FUNC'
(for an external array function / array-valued transformational
intrinsic),  SO WILL FAIL IN THAT CASE !!!
*****************************/
/*** THIS DOESN'T TEST WHETHER src_mapp HAS dim 'src_dim' *******/
/****** DOESN'T COPE WITH src_ep HAVING INHERITED MAPPING! ******/
{
	/*{{{  decls*/
	bool    distrib;
	int32   src_attrs;
	int     dim,  src_embed_rank;
	short  *dp;
	Expr   *array_ep,  **epp,  *n_vec;
	Map    *src_mapp,  *vec_mapp;
	Sect   *vec_sectp;
	/*}}}*/

	/*--------------------------------------*/
	/* Find 'n_vec', the length of 'vec_ep'	*/
	/*--------------------------------------*/
	array_ep = find_array_term (vec_ep);

	if ( ! array_ep->mapp)
		attach_map (array_ep->mapp,  array_ep->val.sp->mapp);
	evaluate_subscrs_and_install_sect (array_ep, NULL_Map);
	n_vec = vec_len (array_ep);
	
	/*--------------------------------------*/
	/* Create and initialise 'vec_mapp'	*/
	/*--------------------------------------*/
	src_mapp = src_ep->mapp;		/* != NULL */
	/*{{{  find 'src_attrs' & 'src_embed_rank'*/
	src_attrs = src_mapp->attrs;
	
	src_embed_rank = ((src_attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	/*}}}*/

	vec_mapp = (Map *) ecalloc (sizeof (Map));

	vec_mapp->nlinks = 0;		/* ...not attached to anything yet! */

	/*{{{  (al,an,as, tn, b, fmt,d) = (0, n_vec, 1, n_vec, b', fmt',d)*/
	/*----------------------------------------------------------------
	(al,an,as, tn, b, fmt,d)  =  (0, n_vec, 1, n_vec, b', fmt', d),
	where:
		n_vec = vec_len (array_ep);
		fmt'  = fmt (src_ep [src_dim]),  except that CYCLIC -> BLOCK
		b'    = { n_vec			-- if (fmt' == COLLAPSED)
		        { CEILING (n_vec / np)	-- if (fmt' == BLOCK)
	----------------------------------------------------------------*/
	vec_mapp->al =   epp = (Expr **) emalloc (5 * sizeof (Expr *));
	vec_mapp->an = ++epp;
	vec_mapp->as = ++epp;
	vec_mapp->tn = ++epp;
	vec_mapp->b  = ++epp;
	
	vec_mapp->fmt =   dp = (short *) emalloc (2 * sizeof (short));
	vec_mapp->d   = ++dp;
	
	attach_expr (vec_mapp->al [0],  int_lit_0);
	attach_expr (vec_mapp->an [0],  n_vec);
	attach_expr (vec_mapp->as [0],  int_lit_1);
	attach_expr (vec_mapp->tn [0],  n_vec);
	
	vec_mapp->d [0] = src_mapp->d [src_dim];	/*** WHAT IF == -2 ?? ******/
	
	switch (src_mapp->fmt [src_dim]) {
	case  DFMT_COLLAPSED :
		vec_mapp->fmt [0] = DFMT_COLLAPSED;
		attach_expr (vec_mapp->b [0],  n_vec);
	
		distrib = FALSE;
		break;
	
	case  DFMT_BLOCK :
	case  DFMT_CYCLIC :
	{
		/*******************************************/
		/****  WHAT IF INHERITED PROCS ???  ********/
		/*******************************************/
		int    procdim = proc_dim (src_mapp->fmt, src_dim);
		Expr  *n_procs = src_mapp->proc_sp->arp->n [procdim];
	
		vec_mapp->fmt [0] = DFMT_BLOCK;
		attach_expr (vec_mapp->b [0], ceiling_div (n_vec, n_procs));
	
		distrib = TRUE;
		break;
	}
	case  DFMT_INHERIT :
		/***********************************/
		/****  WHAT TO DO HERE ???  ********/
		/***********************************/
		distrib = TRUE;
		break;
	}
	/*}}}*/

	if (src_embed_rank) {
		/*{{{  copy (ep, ed) from 'src_mapp' to 'vec_mapp'*/
		/*------------------------------------------------------*/
		/* Copy the original embedded dimensions to 'vec_mapp'.	*/
		/*------------------------------------------------------*/
		short  *edp,   *vec_edp;
		Expr  **vec_epp;
		
		vec_epp = vec_mapp->ep = (Expr **) emalloc (src_embed_rank * sizeof (Expr *));
		vec_edp = vec_mapp->ed = (short *) emalloc (src_embed_rank * sizeof (short));
		
		epp = src_mapp->ep;
		edp = src_mapp->ed;
		
		for (dim = src_embed_rank;    dim;
		                            --dim,  ++epp,  ++edp,  ++vec_epp,  ++vec_edp) {
			attach_expr (*vec_epp,  *epp);
			*vec_edp = *edp;
		}
		/*}}}*/
	}
	vec_mapp->proc_sp = src_mapp->proc_sp;

	/*{{{  set 'vec_mapp->attrs'*/
	vec_mapp->attrs =  MA_INTEGER |
	                  (1 << MA_ARRAY_RANK_SHIFT) |
	                  (src_embed_rank << MA_EMBED_RANK_SHIFT) |
	                  (src_attrs & MA_INHERIT_MASK);
	
	if ( distrib ||  src_embed_rank  ||
	    (src_attrs & (MA_INHERITED | MA_ALIGNED_WITH_INHERIT | MA_INHERIT_FMT)) )
	
		vec_mapp->attrs |= MA_DISTRIB;
	
	else if (vec_mapp->proc_sp  ||  (src_attrs & MA_INHERIT_PROCS))
	
		vec_mapp->attrs |= MA_REPLIC_ON_PROCS;
	
	else
		vec_mapp->attrs |= MA_GLOBALLY_REPLIC;
	/*}}}*/

	/*{{{  create and initialise 'vec_sectp' (to select *whole* of 'vec_mapp')--OK*/
	/*--------------------------------------------------------------*/
	/* Create and initialise 'new_sectp' (to select the *whole* of	*/
	/* the array described by 'new_mapp').				*/
	/*--------------------------------------------------------------*/
	vec_sectp = whole_object_sect (vec_mapp, NULL_Sym);
	/*}}}*/

	*vec_mapp_p  = vec_mapp;
	*vec_sectp_p = vec_sectp;
	return;
}

/*}}}*/
/*{{{  choose_map_for_expr (mapp_p, sectp_p, ep)*/
choose_map_for_expr (mapp_p, sectp_p, ep)
/*-------------------------------------*/
	Map   **mapp_p;		/* result parameter */
	Sect  **sectp_p;	/*   "       "      */
	Expr   *ep;		/* a general expr */
/*----------------------------------------------------------------
N.B.:
--  If this returns 'Map' and 'Sect' structures, they are always
*newly* created ones whose (nlinks == 0), so they can be modified
safely  (e.g. by 'collapse_map_on_p0 ()' or 'replicate_map ()').
-- We can assume that 'sp's mapping *hasn't* been setup, since
this wouldn't be called if it has.

*********************************************
ACTUALLY THIS LAST CLAIM IS WRONG FOR ARG 1 OF THE INTRINSIC REDUCTION
FUNCS IF THEY HAVE A DIM ARG, E.G.:
	PRINT *, any (a, 2)
SINCE THEN ITS MAPPING IS 'CHOOSEN' TWICE.  HOWEVER, IT ALL SEEMS
TO WORK OK ANYWAY!
**********************************************
----------------------------------------------------------------*/
{
	Expr  *array_ep;

	if (ep->type & EX_SCALAR) {
		*mapp_p  = NULL_Map;
		*sectp_p = NULL_Sect;
		return;
	}

	array_ep = find_array_term (ep);  /* ...a ND_ARRAY_SECT /_WHOLE_ARRAY 
					     / FUNC node. */
	if (array_ep->token == ND_FUNC) {
		/*{{{  find a natural (mapp, sectp) for the function result*/
		/*--------------------------------------------------------------*/
		/* 'array_ep' is an array-valued function (either external or	*/
		/* transformational intrinsic -- *not* elemental).  Find a	*/
		/* 'natural' (mapp, sectp) for its result.			*/
		/*								*/
		/* N.B. Since the function is array-valued, it *must* have an	*/
		/* explicit interface.						*/
		/*--------------------------------------------------------------*/
		Symbol  *result_sp = array_ep->val.sp->info.sp;	/* the function result var
								   (in its interface defn.) */
		
		if ( ! (result_sp->attrs & AT_INTRINSIC)) {
			/*----------------------*/
			/* An external function */
			/*----------------------*/
			*mapp_p  = result_sp->mapp;
			*sectp_p = whole_object_sect (result_sp->mapp, result_sp);
			return;
		}
		else if (result_sp->info.fn) {
			result_sp->info.fn (array_ep, mapp_p, sectp_p, TRUE);
			return;
		}
		else
			return;
		/****************************
		!!! NOT GOOD ENOUGH !!!
		WE MUST ENSURE THAT 'result_sp->info_fn ()' IS DEFINED FOR *ALL*
		NON-ELEMENTAL ARRAY-VALUED INTRINSICS!!
		***************/
		/*}}}*/
	}
	else {
		/*----------------------------------------------*/
		/* 'array_ep' is a 'ND_WHOLE_ARRAY /_ARRAY_SECT'*/
		/*----------------------------------------------*/
		attach_map (array_ep->mapp,  array_ep->val.sp->mapp);

		evaluate_subscrs_and_install_sect (array_ep, NULL_Map);

		compose_map_and_sect (mapp_p, sectp_p, array_ep);
			/* creates and returns *new* structures for a
			  'rationalised' mapping equiv to that of 'array_ep' */
	}
}
/*}}}*/

/*{{{  Map  *equiv_scalar_map (mapp) -- OK*/
Map  *equiv_scalar_map (mapp)
/*-------------------------*/
	Map  *mapp;			/* may be NULL */
/*----------------------------------------------------------------
	Called by 'write_remap ()' to create and return a new
'Map' structure, 'new_mapp', that has the same embedding as
'mapp' but omits its array dimensions (or more precisely has its
*distributed* array dimensions converted to replicated ones).
If there are no embedded dimensions, but only replicated ones,
return NULL_Map.
	This is the appropriate mapping for a scalar that is
assigned to an array with mapping 'mapp', i.e. for 'scalar' in
the assignment:
	array = scalar
where 'array' is a 'ND_WHOLE_ARRAY' with mapping 'mapp'.
----------------------------------------------------------------*/
{
	int    dim,  embed_rank;
	short  *dp;
	Map   *new_mapp;

	if (! mapp)
		return NULL_Map;	/* i.e.globally replicated scalar */

	embed_rank = ((mapp->attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	if (! embed_rank)
		return  NULL_Map;

	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	new_mapp = (Map *) ecalloc (sizeof (Map));

	new_mapp->attrs = (mapp->attrs & ~MA_ARRAY_RANK_MASK);

		/* N.B. 'MA_DISTRIB' flag is set as there are embedded dims */

	new_mapp->nlinks = 0;	/* since not attached to anything yet! */

	if (embed_rank) {
		/*{{{  copy (ep, ed) from 'mapp' to 'new_mapp'*/
		/*------------------------------------------------------*/
		/* Copy the original embedded dimensions to 'new_mapp'.	*/
		/*------------------------------------------------------*/
		short  *edp,  *new_edp;
		Expr  **epp,  **new_epp;
		
		new_epp = new_mapp->ep = (Expr **) emalloc (embed_rank * sizeof (Expr *));
		new_edp = new_mapp->ed = (short *) emalloc (embed_rank * sizeof (short));
		
		epp = mapp->ep;
		edp = mapp->ed;
		
		for (dim = embed_rank;    dim;
		                             --dim,  ++epp,  ++edp,  ++new_epp,  ++new_edp) {
			attach_expr (*new_epp,  *epp);
			*new_edp = *edp;
		}
		/*}}}*/
	}

	new_mapp->proc_sp = mapp->proc_sp;

	return  new_mapp;
}
/*}}}*/
/*{{{  Map  *scalar_map_on_p0 () -- ok*/
Map  *scalar_map_on_p0 ()
/*---------------------*/
/*----------------------------------------------------------------
	Called by 'transform_and_write_IO_stmt ()' (directly or
indirectly) to create a 'Map' structure for a scalar embedded
on scalar processor arrangement "$p0", and return a pointer to it.
----------------------------------------------------------------*/
{
	Map  *mapp;
	
	mapp = (Map *) ecalloc (sizeof (Map));
	mapp->attrs   = MA_REPLIC_ON_PROCS;
	mapp->proc_sp = default_proc_sp [0];		/* i.e. "$p0"*/

	return  mapp;
}
/*}}}*/

/*{{{  collapse_map_on_p0 (mapp) -- OK, but check*/
collapse_map_on_p0 (mapp)
/*---------------------*/
	Map  *mapp;		/* != NULL */
/*----------------------------------------------------------------
	Called by 'transform_and_write_IO_stmt ()' (directly or
indirectly) to modify the 'Map' structure 'mapp' so that all its
array dimensions are *collapsed* onto target processor arrangement
"$p0" (a scalar processor arrangement), and all embedded dimensions
are omitted.  I.e. the mapping becomes *embedded* on scalar processor
"$p0".
	This is the appropriate mapping for I/O exprs and vars
*** UNTIL DISTRIBUTED I/O ROUTINES ARE AVAILABLE IN ADLIB ***.

N.B.  Since this overwrites 'mapp', it should be newly created and
unused so far, i.e. its 'nlinks' should be 0 and its 'mapvec_ep'
should be NULL.
************************************************************************
*** DELETE THIS WHEN DISTRIBUTED I/O ROUTINES ARE AVAILABLE IN ADLIB ***
************************************************************************
----------------------------------------------------------------*/
{
	int     array_rank,  dim;
	Expr  **alp,  **anp,  **asp,  **tnp,  **bp;
	short  *fmtp,  *dp;

	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	mapp->attrs &= (MA_TYPE | MA_ARRAY_RANK_MASK);	/* i.e. unset all
							   other flags */
	mapp->attrs |= MA_REPLIC_ON_PROCS;     /* ...i.e. replicated on "$p0"*/

	array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);

	if (array_rank) {
		/*{{{  collapse all array dims*/
		/*--------------------------------------------------------------*/
		/* 'Collapse' each dimension, i.e. set its params to:		*/
		/*	(al,an,as,tn,b,fmt,d) = (0,an,1,an,an,COLLAPSED,-1)	*/
		/*--------------------------------------------------------------*/
		alp = mapp->al;
		anp = mapp->an;
		asp = mapp->as;
		tnp = mapp->tn;
		bp  = mapp->b;
		
		fmtp = mapp->fmt;
		dp   = mapp->d;
		
		for (dim = 0;    dim < array_rank;
		               ++dim,  ++alp,  ++anp,  ++asp,  ++tnp,  ++bp,  ++fmtp,  ++dp) {
		
			if (*dp == -1)
				continue;	/* dimension already collapsed */
		
			/* ... else collapse it.  Only '*anp' remains unchanged. */
		
			delete_expr (*alp, DECR_NLINKS);
			delete_expr (*asp, DECR_NLINKS);
			delete_expr (*tnp, DECR_NLINKS);
			delete_expr (*bp,  DECR_NLINKS);
		
			attach_expr (*alp, int_lit_0);
			attach_expr (*asp, int_lit_1);
			attach_expr (*tnp, *anp);
			attach_expr (*bp,  *anp);
		
			*fmtp = DFMT_COLLAPSED;
			*dp   = -1;
		}
		/*}}}*/
	}
	mapp->proc_sp = default_proc_sp [0];		/* i.e. "$p0" */

}
/*}}}*/
/*{{{  replicate_map (mapp) -- ok, but check*/
replicate_map (mapp)
/*----------------*/
	Map  *mapp;		/* != NULL */
/*----------------------------------------------------------------
	Called by 'broadcast_expr ()' to modify the 'Map' structure
'mapp' so that it's globally replicated (i.e. all array dimensions
are collapsed).

N.B.  Since this overwrites 'mapp', it should be newly created and
unused so far, i.e. its 'nlinks' should be 0 and its 'mapvec_ep'
should be NULL.
----------------------------------------------------------------*/
{
	int     array_rank,  dim;
	Expr  **alp,  **anp,  **asp,  **tnp,  **bp;
	short  *fmtp,  *dp;

	/*--------------------------------------*/
	/* Create and initialise 'new_mapp'	*/
	/*--------------------------------------*/
	mapp->attrs &= (MA_TYPE | MA_ARRAY_RANK_MASK);	/* i.e. unset all
							   other flags */
	mapp->attrs |= MA_GLOBALLY_REPLIC;

	array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);

	if (array_rank) {
		/*{{{  collapse all array dims*/
		/*--------------------------------------------------------------*/
		/* 'Collapse' each dimension, i.e. set its params to:		*/
		/*	(al,an,as, tn,b, fmt,d) = (0,an,1, an,an, COLLAPSED,-1)	*/
		/*--------------------------------------------------------------*/
		alp = mapp->al;
		anp = mapp->an;
		asp = mapp->as;
		tnp = mapp->tn;
		bp  = mapp->b;
		
		fmtp = mapp->fmt;
		dp   = mapp->d;
		
		for (dim = 0;    dim < array_rank;
		               ++dim,  ++alp,  ++anp,  ++asp,  ++tnp,  ++bp,  ++fmtp,  ++dp) {
		
			if (*dp == -1)
				continue;	/* dimension already collapsed */
		
			/* ... else collapse it.  Only '*anp' remains unchanged. */
		
			delete_expr (*alp, DECR_NLINKS);
			delete_expr (*asp, DECR_NLINKS);
			delete_expr (*tnp, DECR_NLINKS);
			delete_expr (*bp,  DECR_NLINKS);
		
			attach_expr (*alp, int_lit_0);
			attach_expr (*asp, int_lit_1);
			attach_expr (*tnp, *anp);
			attach_expr (*bp,  *anp);
		
			*fmtp = DFMT_COLLAPSED;
			*dp   = -1;
		}
		/*}}}*/
	}
	mapp->proc_sp = NULL_Sym;	/* since globally replicated, don't
					   specify target processor array */
}
/*}}}*/

/*--------------------------------------*/
/* deleting 'Map' and 'Sect' structures	*/
/*--------------------------------------*/
/*{{{  delete_sect (sectp,  decr_nlinks)*/
delete_sect (sectp,  decr_nlinks)
/*-----------------------------*/
	Sect  *sectp;
	bool   decr_nlinks;
/*--------------------------------------------------------------*/
/*	Called by 'delete_expr ()' and 'delete_expr_node ()' to	*/
/* delete a 'Sect' record.					*/
/*								*/
/* N.B. We assume that (sectp != NULL)!				*/
/*--------------------------------------------------------------*/
{
	int     dim,  sect_rank;
	Expr  **epp;

	if (decr_nlinks)
		sectp->nlinks--;

	if (sectp->nlinks)
		/*----------------------------------------------*/
		/* 'sectp' is still in use, so don't delete it.	*/
		/*----------------------------------------------*/
		return;

	if (sectp->mapvec_ep)
		delete_expr (sectp->mapvec_ep, DECR_NLINKS);

	sect_rank = (sectp->attrs & SE_RANK_MASK);

	if (sect_rank) {
		/*----------------------------------------------*/
		/* Delete the exprs attached to 'sectp->l []',	*/
		/* etc, as well as the arrays themselves.	*/
		/*						*/
		/* N.B. There are logically 3 arrays of length	*/
		/* 'sect_rank' and type (Expr *):		*/
		/*						*/
		/*	sectp->l[], n[], s[].			*/
		/*						*/
		/* However, they're allocated as a single block	*/
		/* of length (3 * sect_rank) attached to	*/
		/* 'sectp->l' -- so we must 'free' them in the	*/
		/* same fashion.				*/
		/*----------------------------------------------*/
		for (epp = sectp->l,  dim = 3*sect_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);

		free  ((char *) sectp->l);
	}

	free  ((char *) sectp);
}

/*}}}*/
/*{{{  delete_map (mapp,  decr_nlinks)*/
delete_map (mapp,  decr_nlinks)
/*---------------------------*/
	Map  *mapp;
	bool  decr_nlinks;
/*--------------------------------------------------------------*/
/*	Called by 'delete_expr ()' and 'delete_expr_node ()' to	*/
/* delete a 'Map' record.					*/
/*								*/
/* N.B. We assume that (mapp != NULL)!				*/
/*--------------------------------------------------------------*/
{
	int     dim,  array_rank,  embed_rank;
	int32   attrs;
	Expr  **epp;

	if (decr_nlinks)
		mapp->nlinks--;

	if (mapp->nlinks)
		/*----------------------------------------------*/
		/* 'mapp' is still in use, so don't delete it.	*/
		/*----------------------------------------------*/
		return;

	if (mapp->mapvec_ep)
		delete_expr (mapp->mapvec_ep, DECR_NLINKS);

	/*{{{  find 'array_rank' & 'embed_rank'*/
	attrs = mapp->attrs;
	
	array_rank  = (attrs & MA_ARRAY_RANK_MASK)  >> MA_ARRAY_RANK_SHIFT;
	embed_rank  = (attrs & MA_EMBED_RANK_MASK)  >> MA_EMBED_RANK_SHIFT;
	/*}}}*/

	if (array_rank) {
		/*{{{  delete mapp->al, an, as, b, tn, fmt, d*/
		/*--------------------------------------------------------------*/
		/* Delete the exprs attached to 'mapp->al []', etc, as well as	*/
		/* the arrays themselves.					*/
		/*								*/
		/* N.B. There are logically 5 arrays of length 'array_rank' and	*/
		/* type (Expr *):						*/
		/*								*/
		/*	mapp->al[], an[], as[], b[] & tn[].			*/
		/*								*/
		/* However, they're allocated as a single block of length	*/
		/* (5 * array_rank) attached to 'mapp->al' -- so we must 'free'	*/
		/* them in the same fashion.  Likewise the 2 arrays of type	*/
		/* (short), 'fmt []' & 'd []', are allocated and freed as a	*/
		/* single array of length (2 * array_rank) attached to		*/
		/* 'alignp->fmt'.						*/
		/*--------------------------------------------------------------*/
		for (epp = mapp->al,  dim = 5*array_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);
		
		free  ((char *) mapp->al);
		free  ((char *) mapp->fmt);
		/*}}}*/
	}
	if (embed_rank) {
		/*----------------------------------------------*/
		/* Delete the exprs attached to 'mapp->ep []'	*/
		/* and the arrays 'ep []' & 'ed []'.		*/
		/*----------------------------------------------*/
		for (epp = mapp->ep,  dim = embed_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);

		free  ((char *) mapp->ep);
		free  ((char *) mapp->ed);
	}

	free  ((char *) mapp);
}
/*}}}*/

/*}}}*/

/*{{{  << transforming exprs and assignments >>  *****/
/*------------------------------------*/
/* transforming exprs and assignments */
/*------------------------------------*/
/*********************************************************
-- Check that 'sp->arp->lb[]' is always defined before calling these!
It seems currently it isn't unless it's explicit in the input decl.

*********************************************************/
/*---------------------------------*/
/* 'transform and write' functions */
/*---------------------------------*/
/*{{{  transform_and_write_assignment (var_e, expr_e, mask_e) -- DONE VEC & ELEM*/
transform_and_write_assignment (var_ep, expr_ep, mask_ep)
/*-----------------------------------------------------*/
	Expr  *var_ep,  *expr_ep,  *mask_ep;
{
/*---------------------------------------------------------------*/
/** really, we should pass 'mask_ep' into all these functions... */
/*---------------------------------------------------------------*/
	bool   must_scatter,  has_dependence;
	Expr  *temp_ep;
	Map   *tgt_mapp;
	Sect  *tgt_sectp;

	type_check_assignment (var_ep, expr_ep);

	if (var_ep->val.sp->mapp)
		attach_map (var_ep->mapp, var_ep->val.sp->mapp);

	if (var_ep->token == ND_ARRAY_SECT  ||  var_ep->token == ND_WHOLE_ARRAY)
		/*----------------------------------------------*/
		/* O/p code, if necessary, for subscript exprs,	*/
		/* and install 'var_ep->sectp'.			*/
		/*----------------------------------------------*/
		evaluate_subscrs_and_install_sect (var_ep, NULL_Map);

	compose_map_and_sect (&tgt_mapp, &tgt_sectp, var_ep);
		/* returns 'rationalised' mapping equiv to that of 'var_ep' */

	if ( (must_scatter = must_call_scatter (var_ep))  ||
	     (has_dependence = occurs_in_expr (var_ep->val.sp, expr_ep)) )
		temp_ep = alloc_var ((var_ep->type & TY_TYPE), tgt_mapp, tgt_sectp);

	if (mask_ep)		/* in a WHERE construct */
		attain_mapping (mask_ep, tgt_mapp, tgt_sectp, RESULT_VAR);

	attain_mapping (expr_ep, tgt_mapp, tgt_sectp, 0);

	if (must_scatter) {
		write_assignment_stmt (temp_ep, expr_ep, mask_ep);
		write_scatter (var_ep, temp_ep);
		/* ... 1st call deletes 'expr_ep'; 2nd deletes 'var_/temp_ep' */
	}
	else if (has_dependence) {
		write_assignment_stmt (temp_ep, expr_ep, mask_ep);
		write_assignment_stmt (var_ep, temp_ep, mask_ep);
		/* ... 1st call deletes 'expr_ep'; 2nd deletes 'var_/temp_ep' */
	}
	else
		write_assignment_stmt (var_ep, expr_ep, mask_ep);
			/* ... this also deletes 'var_ep' and 'expr_ep' */
}
/*}}}*/

/*{{{  transform_and_write_IO_stmt (name, format, arglist, arglist_type, iolist)*/
transform_and_write_IO_stmt (stmt_name, format, arglist, arglist_type, iolist)
/*--------------------------------------------------------------------------*/
	char  *stmt_name;	/* "READ", "WRITE", "PRINT", "OPEN" etc */
	Expr  *format,		/* E.g. for "READ format";   may be NULL */
	      *arglist;		/* E.g. for "READ (arg-list)";  may be NULL.
				   *One* of 'format' or 'arglist' is defined.*/
	int    arglist_type;	/* Defined if (arglist != NULL) */
	Expr  *iolist;		/* An I/O list.  May be NULL */
{
	Expr  *ep;

	/*------------------------------------------------------*/
	/* check, and if necessary remap, the format or arglist	*/
	/* and the I/O list (if any).				*/
	/*------------------------------------------------------*/
	if (format)
		format = check_and_transform_format_expr (format);
	else {
		/*{{{  check and transform the arglist*/
		/*--------------------------------------*/
		/* Check and transform the 'arglist'	*/
		/*--------------------------------------*/
		switch (arglist_type) {
		case  IO_ARGLIST :
			check_and_transform_IO_arglist (stmt_name, arglist);
			break;
		
		case  OPEN_ARGLIST :
			check_and_transform_OPEN_arglist (arglist);
			break;
		
		case  CLOSE_ARGLIST :
			check_and_transform_CLOSE_arglist (arglist);
			break;
		
		case  FILEPOS_ARGLIST :
			check_and_transform_filepos_arglist (stmt_name, arglist);
			break;
		
		case  INQUIRE_ARGLIST :
			check_and_transform_INQUIRE_arglist (arglist);
			break;
		}
		/*}}}*/
	}
	if (iolist) {
		/*{{{  process the I/O list, remapping items where necessary*/
		/*--------------------------------------------------------------*/
		/* Process the I/O list, remapping items where necessary.	*/
		/*--------------------------------------------------------------*/
		
		switch (*stmt_name) {
		case  'R' :
			/*------------------------------------------------------*/
			/* A 'READ' stmt, so this is an *input* list.		*/
			/*	I.e., its items are (or should be!) *variables*.*/
			/* They're probably not stored just on processor "$p0".	*/
			/* Therefore they must be replaced in the input list by	*/
			/* temp vars which are collapsed onto processor "$p0",	*/
			/* and then *after* the input stmt they must be remapped*/
			/* (or copied) back into the original variables.	*/
			/*							*/
			/* N.B. WE SHOULD CHECK THAT THERE ARE VARIABLES, BUT	*/
			/* CURRENTLY DON'T!! ****				*/
			/*------------------------------------------------------*/
			for (ep = iolist;   ep;   ep = ep->right) {
				/*{{{  replace 'item' by a temp var and set it to copy-back*/
				/*--------------------------------------------------------------*/
				/* 'ep->left' is an input item.  Replace it with a temp var	*/
				/* collapsed on processor "$p0", and set it to copy back into	*/
				/* the original variable *after* the output stmt.		*/
				/*--------------------------------------------------------------*/
				int    datatype;
				Expr  *item_ep,  *temp_ep;
				Map   *tgt_mapp;
				Sect  *tgt_sectp;
				
				item_ep  = ep->left;
				datatype = (item_ep->type & TY_BASIC_TYPE);
				
				if (item_ep->type & EX_SCALAR)
					temp_ep = alloc_var (datatype, scalar_map_on_p0 (), NULL_Sect);
				else {
					choose_map_for_expr (&tgt_mapp, &tgt_sectp, item_ep);
					collapse_map_on_p0 (tgt_mapp);      /* ...modifies 'tgt_mapp';
									       'tgt_sectp' needn't be changed*/
					temp_ep = alloc_var (datatype, tgt_mapp, tgt_sectp);
				}
				replace_expr_by_temp (item_ep, temp_ep, COPY_BACK_IF_VAR);
				/*}}}*/
			}
			break;
		
		default :
			/*------------------------------------------------------*/
			/* A 'WRITE' or 'PRINT' stmt, so this is an *output*	*/
			/* list.						*/
			/*	I.e., its items are *exprs* which may have to	*/
			/* be remapped so that they're collapsed onto processor	*/
			/* "$p0" *before* the output stmt, but don't have to be	*/
			/* copied back afterwards.				*/
			/*------------------------------------------------------*/
			for (ep = iolist;   ep;   ep = ep->right)
				collapse_expr_on_p0 (ep->left, 0);	/* an output item */
			break;
		}
		/*}}}*/
	}

	/*--------------------------------------*/
	/* output the transformed I/O stmt	*/
	/*--------------------------------------*/
	stmt ("IF (K_P0 == 0) THEN");
	increase_indent;
	
	begin_stmt ("%s ", stmt_name);
	if (format)	cont_stmt ("%e",   format);
	else		cont_stmt ("(%e)", arglist);

	if (iolist) {
		if (format) cont_stmt (",");
		cont_stmt (" %e",  iolist);		/* the I/O list */
	}
	end_stmt ("");
	
	reduce_indent;
	stmt ("ENDIF");

	/*------------------------*/
	/* delete the expr(list)s */
	/*------------------------*/
	if (iolist)	delete_expr (iolist, LEAVE_NLINKS);
	if (format)	delete_expr (format, LEAVE_NLINKS);
	else		delete_expr (arglist, LEAVE_NLINKS);
}
/*}}}*/


/*----------------------*/
/* remapping operations */
/*----------------------*/
/*{{{  Expr *attain_mapping (ep, tgt_mapp, tgt_sectp, result_flags)--DONE RESULT_MAPVEC*/
Expr *attain_mapping (ep, tgt_mapp, tgt_sectp, result_flags)
/*--------------------------------------------------------*/
	Expr  *ep;			/* a general expression! */
	Map   *tgt_mapp;		/* NULL for a replicated scalar */
	Sect  *tgt_sectp;		/* NULL for any scalar */
	int    result_flags;		/* e.g. RESULT_VAR, COPY_BACK_IF_VAR */
/*----------------------------------------------------------------
	Called by 'transform_and_write_assignment / IO ()',
'evaluate_subscrs_and_install_sect ()', and also recursively,
to output code if necessary to remap the primaries in 'ep' so that
they're comapped with (tgt_sectp, tgt_mapp).  If 'ep' is indeed
remapped, its identity is *CHANGED* to that of the new remapped expr.
(For this reason it's important to only call this for exprs
appearing explicitly in the input, not for *persistent* exprs stored
internally within ADAPT, e.g. exprs attached to the symtab).
	The function returns a pointer to the remapped expr, which
is always argument 'ep' itself.  It's only provided for ease of use,
e.g. so this can be called within a call to 'attach_expr ()'.
	This function also attaches a 'mapp' record to all
primaries of expr 'ep' (except for globally replicated scalars),
and creates and attaches a 'sectp' record to all array-valued primaries.

**** N.B.  THE LAST ACTION MUST BE PERFORMED BEFORE OUTPUTTING *ANY*
EXPRESSION, SO THIS FUNCTION MUST BE APPLIED TO *ALL* EXPRESSIONS
PRIOR TO OUTPUT! ****
----------------------------------------------------------------*/
/*{{{  << COMMENT -- 'ep->token' values that are ignored >>*/
/*----------------------------------------------------------------
-- Ignore these, since either they're already globally replicated
scalars (whose mapping is always ok), or they don't involve data at all...

ND_CHAR_LIT_CONST
ND_COMPLEX_LIT_CONST
ND_DP_LIT_CONST
ND_INT_LIT_CONST
ND_LOGICAL_LIT_CONST
ND_REAL_LIT_CONST

ND_PROC_ACTUAL_ARG

ND_ALT_RETURN_LABEL

ND_SUBR

ND_INTERN_OBJECT
ND_INTERN_FUNC

ND_STAR
ND_ERROR

-- Ignore these, as we don't expect 'ep' to be one of these.
(In fact, this function checks for these nodes in the subscript
list for 'case ND_ARRAY_SECT', and handles them internally).

ND_EXPRLIST
ND_SUBSCR_DOUBLET
ND_SUBSCR_TRIPLET
----------------------------------------------------------------*/
/*}}}*/
{
	if (!ep)
		return  NULL_Expr;

	if (ep->token & ND_OPER_MASK) {
		/*{{{  attain mapping of operands, and assign to var if necessary*/
		/*--------------------------------------------------------------*/
		/* 'ep' is an operator node.					*/
		/*	Attain the required mapping for the operands, and if	*/
		/* 'result_flags' has flag 'RESULT_VAR' set, allocate a		*/
		/* temporary result variable and assign the expression result	*/
		/* to it.							*/
		/*--------------------------------------------------------------*/
		bool   result_mustbe_var = (result_flags & RESULT_VAR);
		Expr  *temp_ep;
		
		if (result_mustbe_var)
			temp_ep = alloc_var ((ep->type & TY_BASIC_TYPE), tgt_mapp, tgt_sectp);
		
		if (ep->left)
			attain_mapping (ep->left, tgt_mapp, tgt_sectp, 0);
		attain_mapping (ep->right, tgt_mapp, tgt_sectp, 0);
		
		if (result_mustbe_var) {
			ep->nlinks++;		/* ...to 'protect' 'ep' from deletion by
						   the next call... */
		
			write_assignment_stmt (temp_ep, ep, NULL_Expr);
			ep->nlinks--;		/* to 'undo' the earlier protection */
		
			replace_expr_by_temp (ep, temp_ep, (result_flags & RESULT_MAPVEC));
		}
		
		return  ep;
		/*}}}*/
	}

	switch (ep->token) {
	/*{{{  case  ND_SIMPLE_SCALAR : ND_TEMP_SCALAR :*/
	case  ND_SIMPLE_SCALAR :
		/*------------------------------------------------
		If (sp->mapp != NULL), the simple scalar is *embedded*
		in the processor array and so may have to be remapped,
		otherwise it's globally replicated and so is comapped
		with *any* target mapping!
		------------------------------------------------*/
		if ( ! ep->mapp  &&  ep->val.sp->mapp)
			attach_map (ep->mapp, ep->val.sp->mapp);
					/* ...now fall through to next case */
	
	case  ND_TEMP_SCALAR :		/* this already has 'ep->mapp' set */
	
		if (ep->mapp)
			remap_if_necessary (ep, tgt_mapp, tgt_sectp, result_flags);
		break;
	/*}}}*/
	/*{{{  case  ND_ARRAY_SECT : ND_WHOLE_ARRAY : ND_TEMP_ARRAY*/
	case  ND_ARRAY_SECT :
	case  ND_WHOLE_ARRAY :
	
		if ( ! ep->mapp)
			attach_map (ep->mapp, ep->val.sp->mapp);
					/* ...now fall through to next case */
	
	case  ND_TEMP_ARRAY :		/* already has 'ep->mapp/sectp' set */
		/*------------------------------------------------------*/
		/* 'ND_TEMP_ARRAY' can arise due to a previously output	*/
		/* expr being re-used and re-output.  This may happen,	*/
		/* e.g. for a 'mask' expr in a WHERE or FORALL		*/
		/* construct.						*/
		/*------------------------------------------------------*/
		evaluate_subscrs_and_install_sect (ep, tgt_mapp);
				/* remaps subscripts if necessary & sets 'ep->sectp'*/
	
		remap_if_necessary (ep, tgt_mapp, tgt_sectp, result_flags);
		break;
	/*}}}*/
	/*{{{  case  ND_FUNC : ND_SUBR :*/
	case  ND_FUNC :
	case  ND_SUBR :
	/**************************************************
	N.B. If an (array) function result needs to be remapped, ideally
	we should allocate a temp for the remapped result (with mapping
	tgt_mapp, tgt_sectp) *before* allocating anything else.
	Really we need a function 'need_remap ()' so we can predict this!
	**************************************************/
	{
		bool     func_ref = (ep->token == ND_FUNC);
		Expr    *actual_ep,  *arglist,  *result_ep;
		Symbol  *proc_sp   = ep->val.sp,
		        *result_sp = proc_sp->info.sp;	/* the function result var
		        				   or subroutine name;  NULL
							   if no expl interface */
	
		/*------------------------------------------------------*/
		/* Allocate result var and remap args as necessary	*/
		/*------------------------------------------------------*/
		if ( ! (proc_sp->attrs & AT_INFO)) {
			/*{{{  external proc with implicit interface*/
			/*------------------------------------------------------*/
			/* An external procedure with an *implicit* interface	*/
			/*------------------------------------------------------*/
			Map   *actual_mapp;
			Sect  *actual_sectp;
			
			/*--------------------------------------------------------------*/
			/* If this is a function ref, allocate a scalar temporary	*/
			/* variable for the result.					*/
			/*--------------------------------------------------------------*/
			result_ep = func_ref ?
			             alloc_var ((proc_sp->type & TY_BASIC_TYPE), NULL_Map, NULL_Sect) :
			             NULL_Expr;
			
			/*--------------------------------------------------------------*/
			/* Attain a suitable mapping for the actual arguments and	*/
			/* setup their 'mapvec's.					*/
			/*--------------------------------------------------------------*/
			for (arglist = ep->right;   arglist;   arglist = arglist->right) {
			
				actual_ep = arglist->left;
			
				if (actual_ep->token == ND_ALT_RETURN_LABEL  ||
				    actual_ep->token == ND_PROC_ACTUAL_ARG)
					continue;
			
				if (actual_ep->type & EX_SCALAR) {
					attain_mapping (actual_ep, NULL_Map, NULL_Sect, COPY_BACK_IF_VAR);
					continue;
				}
			
				/*------------------------------------------------------*/
				/* ...else an array-valued argument.  Find a suitable	*/
				/* mapping and remap all primaries to it.		*/
				/*------------------------------------------------------*/
				choose_map_for_expr (&actual_mapp, &actual_sectp, actual_ep);
			
				attain_mapping (actual_ep, actual_mapp, actual_sectp,
				                         (COPY_BACK_IF_VAR | RESULT_VAR_AND_MAPVEC));
			
				/*------------------------------------------------------*/
				/* Try to delete 'actual_mapp / sectp' in case they're	*/
				/* unused.						*/
				/*------------------------------------------------------*/
				if (actual_mapp)  delete_map  (actual_mapp,  LEAVE_NLINKS);
				if (actual_sectp) delete_sect (actual_sectp, LEAVE_NLINKS);
			}
			
			write_proc_ref (ep, result_ep);
			/*}}}*/
		}
		else if ( ! (result_sp->attrs & AT_INTRINSIC)) {
			/*{{{  an external proc with an explicit interface*/
			/*------------------------------------------------------*/
			/* An external procedure with an *explicit* interface	*/
			/*------------------------------------------------------*/
			int      arg_flags;
			Map     *actual_mapp;
			Sect    *actual_sectp;
			/*** Sect    *dummy_sectp; -- FOR THE 'PROPER' VERSION! ****/
			Symbol  *dummy_sp,
			        *first_dummy_sp = result_sp->nextid;  /* may be NULL or not a dummy arg */
			
			/*--------------------------------------------------------------*/
			/* If this is a function ref, allocate a temporary variable	*/
			/* (scalar or array) for the result, and allocate and		*/
			/* initialise its 'mapvec'.					*/
			
			/*** BUG:  THIS GIVES NONSENSE IF ANY PARAMETER IN MAPPING OF
			RESULT INVOLVES DUMMY ARGS (OR COMMON BLOCK VARS!!) *****/
			/*--------------------------------------------------------------*/
			if (func_ref) {
				result_ep = alloc_var ((result_sp->type & TY_BASIC_TYPE),
				                        result_sp->mapp, 
				                        whole_object_sect (result_sp->mapp, result_sp));
				setup_mapvec (result_ep);
			}
			else
				result_ep = NULL_Expr;
			
			
			/*{{{  attain the required mapping for the actual args & setup their 'mapvec's*/
			/*--------------------------------------------------------------*/
			/* Attain the required mapping for the actual arguments, and	*/
			/* setup a 'mapvec' for each array-valued argument.		*/
			/*								*/
			/* N.B.:  The argument list has already been checked by		*/
			/* 'proc_ref_primary ()'.  *** We assume here that it's OK,	*/
			/* i.e. that all arguments are correctly associated! ***	*/
			/*--------------------------------------------------------------*/
			for (arglist = ep->right,  dummy_sp = first_dummy_sp;
			     arglist;
			     arglist = arglist->right,  dummy_sp = dummy_sp->nextid) {
			
				actual_ep = arglist->left;
			
				if (actual_ep->token == ND_ALT_RETURN_LABEL  ||
				    actual_ep->token == ND_PROC_ACTUAL_ARG)
					continue;
			
				if (arglist->val.sp) {
					/*----------------------------------------------*/
					/* A keyword argument -- find the corresponding	*/
					/* 'dummy_sp'.  **We assume that the keyword	*/
					/* *is* the name of a dummy argument.**		*/
					/*----------------------------------------------*/
					char  *dummy_name = arglist->val.sp->name;
			
					for (dummy_sp = first_dummy_sp;
					     str_mismatch (dummy_sp->name, dummy_name);
					     dummy_sp = dummy_sp->nextid)
						;
				}
				arg_flags = (dummy_sp->arp ? RESULT_VAR_AND_MAPVEC : 0);
				if ( ! (dummy_sp->attrs & AT_IN)  ||  (dummy_sp->attrs & AT_OUT) )
					arg_flags |= COPY_BACK_IF_VAR;
			
				if (actual_ep->type & EX_SCALAR) {
					attain_mapping (actual_ep, NULL_Map, NULL_Sect, arg_flags);
					continue;
				}
				/*------------------------------------------------------*/
				/* ...else an array-valued argument.  Find a suitable	*/
				/* mapping and remap all primaries to it.		*/
				/*------------------------------------------------------*/
				choose_map_for_expr (&actual_mapp, &actual_sectp, actual_ep);
				attain_mapping (actual_ep, actual_mapp, actual_sectp, arg_flags);
			
				if ((actual_ep->token == ND_ARRAY_SECT)  &&  dummy_sp->arp  &&
				    ! (dummy_sp->arp->attrs & EXPL_SHAPE)) {
			
					/*{{{  ensure actual array sect has no scalar subscripts!*/
					/*--------------------------------------------------------------*/
					/* The dummy arg array is either assumed shape or assumed size.	*/
					/* Hence it derives its dimensions from the dummy 'mapvec'.	*/
					/* For this to work, it's essential that the actual arg has	*/
					/* the same no. of dims as the dummy, i.e. the actual mustn't	*/
					/* have any scalar subscripts.					*/
					/*	If it does, copy it into a temporary with the same	*/
					/* mapping.							*/
					/*--------------------------------------------------------------*/
					bool    must_copy;
					int     array_rank,  dim;
					Expr  **np;
					
					
					array_rank = ((actual_mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
					np = actual_ep->sectp->n;
					must_copy = FALSE;
					
					for (dim = array_rank;   dim;   --dim,  ++np) {
						if (*np == int_lit_m1) {
							must_copy = TRUE;
							break;
						}
					}
					
					if (must_copy) {
						Expr  *temp_ep;
					
						temp_ep = alloc_var ((actual_ep->type & EX_BASIC_TYPE),
						                      actual_mapp, actual_sectp);
						setup_mapvec (temp_ep);
					
						protect_expr (actual_ep);	/* ...to prevent the following
										   from deleting it! */
						write_assignment_stmt (temp_ep, actual_ep, NULL_Expr);
					
						actual_ep->nlinks--;		/* to remove earlier protection */
					
					/***********************************************************
					NOW WE MUST DELETE 'actual_ep->sectp', AS IT'S BEEN LOCALISED.
					WHEN COPY-BACK OCCURS IT WILL BE RE-CREATED.
					(THIS MAY BE A BIT DUBIOUS -- NEED TO CHECK CAREFULLY!)
					**************************************************************/
						delete_sect (actual_ep->sectp, DECR_NLINKS);
						actual_ep->sectp = NULL_Sect;
					
						replace_expr_by_temp (actual_ep, temp_ep, COPY_BACK_IF_VAR);
					}
					/*}}}*/
				}
				/*------------------------------------------------------*/
				/* Try to delete 'actual_mapp / sectp' in case they're	*/
				/* unused.						*/
				/*------------------------------------------------------*/
				if (actual_mapp)  delete_map  (actual_mapp,  LEAVE_NLINKS);
				if (actual_sectp) delete_sect (actual_sectp, LEAVE_NLINKS);
			
				/*{{{  **** THE 'PROPER' VERSION -- ONCE WE'VE SORTED IT OUT!! *****/
				/**************************************************************
				*************************************************************
				-- THIS VERSION TAKES THE MAPPING FROM THE DUMMY ARG SPEC (AS WE SHOULD!).
				HOWEVER, THERE'S A PROBLEM, IN THAT THE MAPPING CAN CONTAIN SPEC EXPRS
				WHICH DEPEND ON DUMMY ARGS  ( & COMMON BLOCK VARS??).
				IN THAT CASE WE MUST 'BIND' THE APPROPRIATE ACTUAL ARG VALUES
				BEFORE COMPUTING THE ACTUAL'S MAPPING!!
				*************************************************************
				
				if (dummy_sp->mapp  &&  (dummy_sp->mapp->attrs & MA_ASSUMED_SHAPE)) {
					**--------------------------------------------------------------**
					** The dummy arg has assumed shape so its 'mapp' is meaningless,**
					** so ignore it.  Just attain a 'natural' mapping for the	**
					** actual arg, as though the interface were implicit.		**
					**--------------------------------------------------------------**
					Map   *actual_mapp;
					Sect  *actual_sectp;
					
					choose_map_for_expr (&actual_mapp, &actual_sectp, actual_ep);
					
					attain_mapping (actual_ep, actual_mapp, actual_sectp, arg_flags);
					
					if (actual_mapp)  delete_map  (actual_mapp,  LEAVE_NLINKS);
					if (actual_sectp) delete_sect (actual_sectp, LEAVE_NLINKS);
				}
				else {
					dummy_sectp = whole_object_sect (dummy_sp->mapp, dummy_sp);
				
					attain_mapping (actual_ep, dummy_sp->mapp, dummy_sectp,  arg_flags);
					if (dummy_sectp)
						delete_sect (dummy_sectp, LEAVE_NLINKS);
				}
				**************************************************************/
				/*}}}*/
			}
			/*}}}*/
			
			write_proc_ref (ep, result_ep);		/* N.B. This doesn't delete the
								   result's 'mapvec' (if it has one)*/
			if (func_ref  &&  ep->mapp)
				remap_if_necessary (ep, tgt_mapp, tgt_sectp,
				                               (result_flags & RESULT_MAPVEC));
			/*}}}*/
		}
		else if (result_sp->type & TY_ELEMENTAL) {
			if (func_ref) {
				/***************************************
				CHECK THAT THIS IS WRITTEN OUT OK.
				E.G. WHAT ABOUT MISSING OPTIONAL ARGS?
				*****************************************/
	
				/*{{{  an elemental intrinsic *function**/
				/*--------------------------------------*/
				/* An elemental intrinsic *function*.	*/
				/*--------------------------------------*/
				/*--------------------------------------------------------------*/
				/* If all of the arguments, and hence the result, are scalar,	*/
				/* change the target mapping to the 'equivalent scalar' mapping.*/
				/*--------------------------------------------------------------*/
				if (ep->type & EX_SCALAR) {
					tgt_mapp  = equiv_scalar_map (tgt_mapp);  /* NULL if no embedded dims*/
					tgt_sectp = NULL_Sect;
				}
				
				/*--------------------------------------------------------------*/
				/* If all of the arguments are scalar, or 'result_flags' has	*/
				/* flag 'RESULT_VAR' set, allocate a temporary variable (scalar	*/
				/* or array) for the result.					*/
				/*--------------------------------------------------------------*/
				if ((ep->type & EX_SCALAR)  ||  (result_flags & RESULT_VAR))
					result_ep = alloc_var ((ep->type & TY_BASIC_TYPE), tgt_mapp, tgt_sectp);
				else
					result_ep = NULL_Expr;
				
				
				/*--------------------------------------------------------------*/
				/* Attain mapping 'tgt_mapp / sectp' for each actual argument.	*/
				/*--------------------------------------------------------------*/
				for (arglist = ep->right;   arglist;   arglist = arglist->right)
					attain_mapping (arglist->left, tgt_mapp, tgt_sectp, 0);
				
				
				/*--------------------------------------------------------------*/
				/* If a new 'tgt_mapp' was created by 'equiv_scalar_map ()',	*/
				/* try to delete it in case it's unused.			*/
				/*--------------------------------------------------------------*/
				if ((ep->type & EX_SCALAR)  &&  tgt_mapp)
					delete_map (tgt_mapp, LEAVE_NLINKS);
				
				if (result_ep) {
					write_assignment_stmt (result_ep, ep, NULL_Expr);
					replace_expr_by_temp (ep, result_ep, (result_flags & RESULT_MAPVEC));
				}
				/*}}}*/
			}
			else {
				not_supported ("intrinsic subroutine %q", proc_sp->name);
	
				/**************************************/
				/**** MVBITS:  GOT TO DO THIS! ********/
				/**************************************/
				/*{{{  an elemental intrinsic *subroutine* (viz. MVBITS!)*/
				/***********************************
				************************************************************************
				STILL GOT TO DO THIS.  ALSO HAVE TO WRITE SOMETHNG SPECIAL TO DO THE OUTPUT!
				***********************************************************************
				**--------------------------------------------------------------**
				** An elemental intrinsic subroutine (which currently can only	**
				** be 'MVBITS ()'!).						**
				**								**
				** N.B. Because this is a subroutine call, 'tgt_mapp / sectp'	**
				** are initially NULL.  They must be set to suit the output	**
				** argument(s) ('TO' in this case!).				**
				**--------------------------------------------------------------**
				Symbol  *dummy_sp,
				        *first_dummy_sp = result_sp->nextid;
				
				**--------------------------------------------------------------**
				** If all of the arguments are scalar, change the target	**
				** mapping to the 'equivalent scalar' mapping.			**
				**--------------------------------------------------------------**
				if (ep->type & EX_SCALAR) {
					tgt_mapp  = equiv_scalar_map (tgt_mapp);  ** NULL if no embedded dims**
					tgt_sectp = NULL_Sect;
				}
				
				**--------------------------------------------------------------**
				** Attain mapping 'tgt_mapp / sectp' for each actual argument.	**
				**								**
				** N.B.:							**
				** --  The argument list has already been checked by		**
				** 'proc_ref_primary ()'.  *** We assume here that it's OK,	**
				** i.e. that all arguments are correctly associated! ***	**
				** -- The only reason for finding the corresponding 'dummy_sp'	**
				** is to set flag 'COPY_BACK_IF_VAR' for the INTENT (INOUT)	**
				** argument of elemental subroutine 'MVBITS()' !!		**
				**--------------------------------------------------------------**
				for (arglist = ep->right,  dummy_sp = first_dummy_sp;
				     arglist;
				     arglist = arglist->right,  dummy_sp = dummy_sp->nextid) {
				
					if (arglist->val.sp) {
						**----------------------------------------------**
						** A keyword argument -- find the corresponding	**
						** 'dummy_sp'.  **We assume that the keyword	**
						** *is* the name of a dummy argument.**		**
						**----------------------------------------------**
						char  *dummy_name = arglist->val.sp->name;
				
						for (dummy_sp = first_dummy_sp;
						     str_mismatch (dummy_sp->name, dummy_name);
						     dummy_sp = dummy_sp->nextid)
							;
					}
				
					attain_mapping (arglist->left, tgt_mapp, tgt_sectp,
					             ((dummy_sp->attrs & AT_OUT) ? COPY_BACK_IF_VAR : 0));
				}
				
				**--------------------------------------------------------------**
				** If a new 'tgt_mapp' was created by 'equiv_scalar_map ()',	**
				** try to delete it in case it's unused.			**
				**--------------------------------------------------------------**
				if ((ep->type & EX_SCALAR)  &&  tgt_mapp)
					delete_map (tgt_mapp, LEAVE_NLINKS);
				
				**************************************
				**** MUST WRITE SOME OUTPUT!! ********
				**************************************
				************************************/
				/*}}}*/
			}
		}
		else if (result_sp->info.fn) {
			/*{{{  an array-valued intrinsic (with a special func for translation)*/
			/*--------------------------------------------------------------*/
			/* An array-valued transformational intrinsic with a special	*/
			/* function, 'result_sp->info.fn ()', defined to transform	*/
			/* and write the reference.					*/
			/*								*/
			/* N.B.  This function always write the result to a temporary	*/
			/* variable and creates an associated 'mapvec' (unless the	*/
			/* result is scalar), so we can forget about 'result_flags'	*/
			/* (i.e. about whether it has (RESULT_VAR | RESULT_MAPVEC) set).*/
			/*--------------------------------------------------------------*/
			result_sp->info.fn (ep, &tgt_mapp, &tgt_sectp, FALSE);
			
			if (func_ref  &&  ep->mapp)
				remap_if_necessary (ep, tgt_mapp, tgt_sectp, 0);
			/*}}}*/
		}
	/********* TRANSFORMATIONAL / INQUIRY ?? *******/
	
		else if (result_sp->attrs & AT_SUBR)
			not_supported ("intrinsic subroutine %q", proc_sp->name);
		else
			not_supported ("intrinsic function %q", proc_sp->name);
	
		break;
	}
	/*}}}*/
	}

	return  ep;
	/******************** FORGET THESE FOR NOW! **********
	case  ND_STRUCT_CONSTR :
	******************************************************/
}
/*}}}*/

/*{{{  remap_if_necessary (src_ep, tgt_mapp, tgt_sectp, result_flags)--DONE RESULT_MAPVEC*/
remap_if_necessary (src_ep, tgt_mapp, tgt_sectp, result_flags)
/*----------------------------------------------------------*/
	Expr  *src_ep;		/* ND_SIMPLE_SCALAR / TEMP_SCALAR / ARRAY_SECT /
				      WHOLE_ARRAY / TEMP_ARRAY.  It has at
				      least 'src_ep->mapp' set (i.e. it
				      isn't a globally replicated scalar) */
	Map   *tgt_mapp;	/* NULL for a globally replicated scalar */
	Sect  *tgt_sectp;	/* NULL for any scalar */
	int    result_flags;	/* ...only 'COPY_BACK_IF_VAR' and
				      'RESULT_MAPVEC' are relevant */
/*----------------------------------------------------------------
	Called by 'attain_mapping ()' to do all the necessary
processing (including generating the output) to make a data object
'src_ep' (a simple scalar, array, array section or array element)
co-distributed with the target mapping 'tgt_mapp'.  'src_ep' has
'mapp' set -- i.e. it isn't a globally replicated simple scalar.

N.B.:
--  'src_ep->sectp' *isn't* 'localised' yet.
--  'tgt_sectp' can be ignored, as it simply selects the *whole*
array described by 'tgt_mapp'!  It's only passed-in to pass-on
to 'alloc_var ()' (via 'write_gather / _remap ()'), which attaches
it to 'temp_ep'.
----------------------------------------------------------------*/
{
	/*{{{  decls*/
	bool     irregular_sect;
	int      src_dim,  src_embed_rank,  src_rank,
	         tgt_dim,  tgt_embed_dim,  tgt_embed_rank,  tgt_rank;
	short   *src_dp,  *src_edp,  *tgt_edp;
	Expr   **src_epp,  **src_np,  **tgt_epp;
	Map     *src_mapp  = src_ep->mapp;	/* != NULL */
	Sect    *src_sectp = src_ep->sectp;	/* NULL for embedded scalar */
	/*}}}*/


	/*{{{  check 'src_ep's rank and whether it's an irregular section*/
	src_rank = ((src_mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	tgt_rank = tgt_mapp ?
	             ((tgt_mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT) :
	             0;
	
	irregular_sect = FALSE;
	if (src_rank) {
		int  src_sect_rank;
	
		src_np = src_sectp->n;	/* 'src_sectp' is defined as 'src' is an array*/
		
		for (src_sect_rank = src_dim = 0;    src_dim < src_rank;
		     ++src_dim,  ++src_np) {
	
			if (*src_np == int_lit_m2)
				/*----------------------*/
				/* an irregular section */
				/*----------------------*/
				irregular_sect = TRUE;
	
			if (*src_np != int_lit_m1)
				/*--------------------------------------*/
				/* a regular or irregular section, so	*/
				/* it contributes to the section's rank.*/
				/*--------------------------------------*/
				++src_sect_rank;
		}
	
		if (src_sect_rank  &&  (src_sect_rank != tgt_rank)) {
			errmsg ("right-hand-side expression does not conform with assignment variable");
			return;		/* forget about remapping in this erroneous case */
		}
	}
	/*}}}*/

	if (src_mapp->attrs & MA_GLOBALLY_REPLIC)
		/*------------------------------------------------
		'src' is globally replicated, so is comapped with *any*
		'tgt' (even one with distributed vector subscripts).
		------------------------------------------------*/
		goto  mapping_ok;

	/*---------------------------------------------------------*/
	/* ...else 'src_ep' is distributed, embedded or replicated */
	/* over a specified (or inherited) processor arrangement.  */
	/*---------------------------------------------------------*/
/****** N.B. THE FOLLOWING ASSUMES THAT 'proc_sp' IS DEFINED IF
'MA_GLOBALLY_REPLIC' ISN'T SET -- I.E. IT IGNORES 'INHERIT_PROCS' ****/

	if ( ! tgt_mapp  ||  (tgt_mapp->attrs & MA_GLOBALLY_REPLIC)  ||
	     ! ((src_mapp->proc_sp == tgt_mapp->proc_sp)  ||
	         shapes_equal (src_mapp->proc_sp->arp, tgt_mapp->proc_sp->arp)) )
		/*------------------------------------------------
		'src' must be remapped, since either it's distributed
		but 'tgt' is globally replicated, or 'src' and 'tgt'
		are distributed over different processor arrangements.
		------------------------------------------------*/
		goto  remap;

	/*------------------------------------------------------*/
	/* ...else 'src' & 'tgt' both distrib/embedded. Must	*/
	/* 'src' be remapped?					*/
	/*------------------------------------------------------*/
	/*{{{  find 'src_ / tgt_embed_rank'*/
	src_embed_rank = ((src_mapp->attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	tgt_embed_rank = ((tgt_mapp->attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT);
	/*}}}*/
	
	/*{{{  if any dim of 'src' isn't co-mapped with 'tgt', goto remap*/
	/*----------------------------------------------------------------
		'src' and 'tgt' are both distributed and/or embedded.
		If any dimension of 'src' isn't co-mapped with 'tgt',
	'src' must be remapped (so "goto remap").  A dimension of 'src'
	is co-mapped with the 'tgt' if:
	
	-- it's collapsed;  OR
	-- it selects one element  &&  is resident on a proc coord that 'tgt' is
	                               embedded in;  OR
	-- it's a regular section  &&  is aligned with corresponding dim of 'tgt'
	                           &&  distributed over same proc array dim.
	
	N.B. If an *irregular* section is selected in a *distributed* dimension,
	then 'AD_GATHER' must always be called regardless of the details of
	the distribution.  We assume that the vector subscript is already
	mapped suitably.
	----------------------------------------------------------------*/
	if (src_rank) {
		tgt_dim = -1;		/* N.B. 'tgt_mapp' selects the *whole* of
					   every dim (not an element or section) */
	
		src_np = src_sectp->n;	/* ...(src_sectp != NULL) as 'src' is an array*/
		src_dp = src_mapp->d;
	
		for (  src_dim = 0;    src_dim < src_rank;
		     ++src_dim,  ++src_np,  ++src_dp) {
	
			if (*src_np != int_lit_m1)
				++tgt_dim;	/*  a 'sectional' dim;  increment
						    the corresponding tgt dim */
			if (*src_dp == -1)
				continue;	/* 'src' dim collapsed, so OK */
	
			/*----------------------------------------------*/
			/* ...else 'src_dim' is distributed.  Is it	*/
			/* co-mapped with 'tgt'?			*/
			/*----------------------------------------------*/
			if (*src_np == int_lit_m1) {
				/*{{{  a single element: (l,n,s) == (i, -1, 0)*/
				/*--------------------------------------------------------------*/
				/* A single element in a distributed dimension:			*/
				/*	(l, n, s) = (i, -1, 0).					*/
				/* If 'tgt' has an embedded dim with parameters (ep, ed) ==	*/
				/* (src_proc, src_dim), then OK;  else must remap.		*/
				/*--------------------------------------------------------------*/
				tgt_epp = tgt_mapp->ep;
				tgt_edp = tgt_mapp->ed;
				
				for (  tgt_embed_dim = 0;    tgt_embed_dim < tgt_embed_rank;
				     ++tgt_embed_dim,  ++tgt_epp,  ++tgt_edp) {
				
					if (*src_dp == *tgt_edp) {
						Expr  *src_proc;
				
						proc_and_subscript (&src_proc,  NULL_Expr_p,
						                    src_mapp, src_sectp->l [src_dim], src_dim);
				
						if (exprs_equal (src_proc, *tgt_epp)) {
							/*--------------------------------------*/
							/* the target is embedded at this proc	*/
							/* coord, so it's mapped OK		*/
							/*--------------------------------------*/
							delete_expr (src_proc, DECR_NLINKS);
							goto  this_dim_ok;
						}
						else {
							delete_expr (src_proc, DECR_NLINKS);
							goto  remap;
						}
					}
				}
				goto  remap;	/* tgt *isn't* embedded at this proc coord */
				/*}}}*/
			}
			else if (*src_np != int_lit_m2) {
				/*{{{  a regular section*/
				/*{{{  << EXPLANATION >>*/
				/*----------------------------------------------------------------
					A regular section in a *distributed* dimension, which must
				correspond with the next target array dim.
					If the 'src' and 'tgt' regular sections are distributed
				over the same process array dimension (i.e. src->d == tgt->d)
				*AND* corresponding elements are aligned, then they're comapped.
				Otherwise 'src' must be remapped.
					The procedure for testing alignment is as follows:
				
				-- For the 'src', combine (l,s) * (al, as) to get an equivalent
				(equiv_al, equiv_as), s.t. the *whole* dimension described by
				the latter has the same extent & mapping as the regular section
				described by the former.  (This has already been done for the
				'tgt_mapp' parameters, by 'compose_map_and_sect ()'.)
				
				-- 'src' and 'tgt' dims are aligned iff:
					(equiv_al, equiv_as, src_b) == (tgt_al, tgt_as, tgt_b)
				
				N.B.:
				-----
				-- Whether (src->fmt == tgt->fmt) is irrelevant.  We know that
				each is either 'DFMT_BLOCK / _CYCLIC / _INHERITED';  whichever they
				are, they're aligned if the given test is passed.
				
				-- Currently we don't test whether (src->n == tgt->an), but instead
				trust that they're equal (though they needn't necessarily be
				identical expressions!).  In fact it's an error if they aren't,
				since then 'src' & 'tgt' don't conform!
					Eventually we should probably check this -- if their
				difference is a (contant != 0), give an errmsg;  if the difference
				isn't a constant, insert a runtime test.
				----------------------------------------------------------------*/
				/*}}}*/
				
				bool   aligned;
				Expr  *equiv_al,  *equiv_as;
				
				if (*src_dp != tgt_mapp->d [tgt_dim])
					goto remap;		/* distributed over different proc array dims */
				
				/*--------------------------------------------------------------*/
				/* Combine 'src's regular section parameters as follows:	*/
				/*	equiv_al  = src_al + (src_l * src_as)			*/
				/*	equiv_as  = (src_s * src_as)				*/
				/****** COULD SIMPLIFY FURTHER BY DIVIDING BY hcf (b, |as)) ********/
				/*--------------------------------------------------------------*/
				equiv_al = expr (src_mapp->al [src_dim],
				                 ND_PLUS_OP,
				                 expr (src_sectp->l [src_dim],
				                       ND_TIMES_OP,
				                       src_mapp->as [src_dim]
				                      )
				                );
				
				equiv_as = expr (src_sectp->s [src_dim],
				                 ND_TIMES_OP,
				                 src_mapp->as [src_dim]
				                );
				
				aligned = (exprs_equal (equiv_al,  tgt_mapp->al [tgt_dim])
				       &&  exprs_equal (equiv_as,  tgt_mapp->as [tgt_dim])
				       &&  exprs_equal (src_mapp->b [src_dim],  tgt_mapp->b [tgt_dim]));
				
				delete_expr (equiv_al, LEAVE_NLINKS);
				delete_expr (equiv_as, LEAVE_NLINKS);
				
				if (aligned)
					goto  this_dim_ok;	/* they're aligned!! */
				else
					goto remap;		/* not aligned! */
				/*}}}*/
			}
			else
				/*---------------------------------------*/
				/* An irregular section in a distributed */
				/* dim, so 'AD_GATHER ()' must be called.*/
				/*---------------------------------------*/
				goto remap;
	this_dim_ok :
			continue;
		}
	}
	/*}}}*/

	/*{{{  if 'src' is embedded at a proc coord that 'tgt' isn't, goto remap*/
	/*----------------------------------------------------------------
		If 'src' is embedded in any processor array dimensions, is
	'tgt' embedded in the same processor dimension at the same
	processor coordinate?  If not, 'src' must be remapped (or
	more precisely broadcast in this processor array dimension).
	----------------------------------------------------------------*/
	src_epp = src_mapp->ep;
	src_edp = src_mapp->ed;
	
	for (  src_dim = 0;    src_dim < src_embed_rank;
	     ++src_dim,  ++src_epp,  ++src_edp) {
	
		/*--------------------------------------------------------------*/
		/* 'src' is embedded in processor array dimension '*src_edp' at	*/
		/* coordinate '*src_epp'.  Is 'tgt' likewise embedded?  If not	*/
		/* 'src' must be remapped (or more precisely broadcast in this	*/
		/* processor array dimension).					*/
		/*--------------------------------------------------------------*/
		tgt_epp = tgt_mapp->ep;
		tgt_edp = tgt_mapp->ed;
		
		for (  tgt_embed_dim = 0;    tgt_embed_dim < tgt_embed_rank;
		     ++tgt_embed_dim,  ++tgt_epp,  ++tgt_edp) {
		
			if (*tgt_edp != *src_edp)
				continue;
	
			else if (exprs_equal (*tgt_epp, *src_epp))
				goto  this_embed_dim_ok;  /* 'tgt' is embedded at
							     this proc coord, so OK */
			else
				goto  remap;		  /* tgt *isn't* embedded at
							     this proc coord */
		}
		goto  remap;	/* tgt *isn't* embedded in this proc dim at all */
	
	this_embed_dim_ok :
			continue;
	}
	
	/*}}}*/

mapping_ok :
	if (result_flags & RESULT_MAPVEC)
		setup_mapvec (src_ep);
	return;
	
remap :
	if (irregular_sect)
		write_gather (src_ep, tgt_mapp, tgt_sectp);
	else
		write_remap (src_ep, tgt_mapp, tgt_sectp, result_flags);
}
/*}}}*/

/*{{{  localise_array_expr (ep, sect_dim, eta_p) -- DONE VEC & ELEM*/
localise_array_expr (ep, sect_dim, eta_p)
/*-------------------------------------*/
	Expr  *ep;		/* != NULL */
	int    sect_dim;	/* 0 unless 'ep' is a vector subscript in
				   dimension 'sect_dim' */
	Expr  **eta_p;		/* I.e. pointer to 'eta []'.  For each
				   section dim, gives the number of the
				   first element in the local segment.
				   If its entries are NULL they're defined
				   here, otherwise they're used here. */
/*--------------------------------------------------------------*/
/*	Called by 'write_assignment_stmt ()', and also		*/
/* recursively.							*/
/*--------------------------------------------------------------*/
{
	if (ep->token & ND_OPER_MASK) {
		if (ep->left)
			localise_array_expr (ep->left, sect_dim, eta_p);
		localise_array_expr (ep->right, sect_dim, eta_p);
	}
	else if (ep->token == ND_FUNC) {
		/*{{{  if it's an elemental function, 'localise' its args*/
		/*--------------------------------------------------------------*/
		/* If 'ep' is an elemental function, 'localise' its args.	*/
		/*--------------------------------------------------------------*/
		Symbol  *sp = ep->val.sp;
		
		if ((sp->attrs & AT_INFO)  &&
		    (sp->info.sp->attrs & AT_INTRINSIC)  &&
		    (sp->info.sp->type & TY_ELEMENTAL)) {
			/*------------------------------*/
			/* An elemental function ref.	*/
			/*------------------------------*/
			Expr  *arg_ep;
		
			arg_ep = ep;
			while (arg_ep = arg_ep->right)	/* 'arg_ep' = a 'ND_EXPRLIST' node */
				localise_array_expr (arg_ep->left, sect_dim, eta_p);
		}
		/*}}}*/
	}
	else if (ep->sectp  &&  ! (ep->sectp->attrs & SE_LOCALISED_SECT)) {
		/*{{{  localise 'sectp'*/
		/*----------------------------------------------------------------
			'ep' is an array-valued object, i.e. a 'ND_WHOLE_ARRAY',
		'ND_ARRAY_SECT', or 'ND_TEMP_ARRAY' node.
			For each dimension 'dim', convert its 'sectp' parameters
		(l,n,s) from parent subscripts -> local subscripts ('localise' them),
		i.e.:
		
		-- regular section dim:  (l, n, s)  ->  (loc_l, loc_n, loc_s)
		-- single element dim:   (i,-1, 0)  ->  (loc_i, -1, 0)
		-- vector subscript dim: (v,-2, 0)  ->  localise_array_expr (v)
		
		If it's a collapsed dimension, then 'parent' subscripts == 'local'
		subscripts so (l,n,s) shouldn't be modified.
			If they aren't already defined, this also sets 'eta [sect_dim]'
		for every regular section dimension 'sect_dim'.
		----------------------------------------------------------------*/
		int     array_rank,  dim;
		short  *dp;
		Expr  **lp,  **np,  **sp;
		Map    *mapp  = ep->mapp;	/* != NULL */
		Sect   *sectp = ep->sectp;	/* != NULL */
		
		array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
		
		lp = sectp->l;
		np = sectp->n;
		sp = sectp->s;
		dp = mapp->d;
		
		for (dim = 0;    dim <= array_rank-1;    ++dim,  ++lp,  ++np,  ++dp) {
		
			if (*np == int_lit_m2) {
				/*{{{  an irregular section; 'localise' vector subscript 'v'*/
				/*--------------------------------------------------------------*/
				/* An irregular section.  'Localise' the vector subscript expr,	*/
				/* viz. 'l' in (l,-2, 0).					*/
				/*--------------------------------------------------------------*/
				localise_array_expr (*lp, sect_dim, eta_p);
				
				++eta_p;
				++sect_dim;
				/*}}}*/
			}
			else if (*np != int_lit_m1) {
				/*{{{  a regular section:  modify (l,n,s) and set '*eta_p' as nec*/
				/*----------------------------------------------------------------
					A regular section.  Modify (l,n,s) and set '*eta_p' as
				necessary.
					Also, replace any of (l,n,s) that isn't an int-lit-const
				or a simple scalar variable by a scalar temporary, to which the
				parameter value is assigned.  This is done for two reasons:
				-- to avoid the possibility of dependence (which occurs if the
				parameter happens to involve the assignment stmt's assignment variable);
				-- optimisation (a beneficial side-effect!).
				----------------------------------------------------------------*/
				if (*dp != -1)
					/*------------------------------------------------------*/
					/* A regular section in a *distributed* dimension.	*/
					/*							*/
					/* N.B. The following call either defines '*eta_p' (for	*/
					/* the assignment variable) or uses it (for a RHS term).*/
					/*------------------------------------------------------*/
					localise_reg_sect_dim (ep, dim, sect_dim, eta_p);
				
				else if (! *eta_p) {
					/*------------------------------------------------------*/
					/* A regular section in a *collapsed* dimension of the	*/
					/* *assignment variable*.				*/
					/*	Don't change (l,n,s) -- see above -- but set	*/
					/* '*eta_p' to 0  (i.e. the 1st local element is # 0).	*/
					/*------------------------------------------------------*/
					attach_expr (*eta_p, int_lit_0);
					/*{{{  save (l,n,s) in scalar temps (for dependence and optimisation)*/
					/*--------------------------------------------------------------*/
					/* Save (l,n,s) in scalar temps (both to avoid the possibility	*/
					/* of dependence and for optimisation).				*/
					/*--------------------------------------------------------------*/
					(*lp)->nlinks--;			/* to 'detach' it */
					attach_expr (*lp, save_scalar_in_temp (*lp));
					
					(*np)->nlinks--;			/* to 'detach' it */
					attach_expr (*np, save_scalar_in_temp (*np));
					
					(*sp)->nlinks--;			/* to 'detach' it */
					attach_expr (*sp, save_scalar_in_temp (*sp));
					
					/*}}}*/
				}
				else if (*eta_p != int_lit_0) {
					/*------------------------------------------------------*/
					/* A regular section in a *collapsed* dimension on the	*/
					/* *RHS*, and the corresponding dimension of the	*/
					/* assignment variable is *distributed*.  Hence replace:*/
					/*	l -> loc_l = l + eta * s			*/
					/* (using the fact that (al,as) == (0,1)).		*/
					/*							*/
					/* N.B. We should also change  n -> loc_n, where 'loc_n'*/
					/* is given by the assignment variable's 'sectp',  but	*/
					/* don't as the latter is used directly.		*/
					/*------------------------------------------------------*/
					Expr  *parent_l = *lp;
				
					attach_expr (*lp,  save_scalar_in_temp (
					                   expr (parent_l,
					                         ND_PLUS_OP,
					                         expr (*eta_p, ND_TIMES_OP, sectp->s [dim]))
					            ));
					delete_expr (parent_l, DECR_NLINKS);
				
					(*sp)->nlinks--;		/* to 'detach' it */
					attach_expr (*sp, save_scalar_in_temp (*sp));
				}
				++eta_p;  ++sect_dim;
				/*}}}*/
			}
			else if (*dp != -1) {
				/*{{{  a single element in a distrib dim;  localise subscript 'i'*/
				/*------------------------------------------------------*/
				/* A single element in a distributed dimension.		*/
				/* sectp = (i, -1,0).  Convert 'i' to local subscripts.	*/
				/*------------------------------------------------------*/
				Expr  *parent_i = *lp;
				
				proc_and_subscript (NULL_Expr_p,  lp,  mapp,  parent_i,  dim);
				
				delete_expr (parent_i, DECR_NLINKS);
				
				(*lp)->nlinks--;		/* to 'detach' it */
				attach_expr (*lp, save_scalar_in_temp (*lp));
				/*}}}*/
			}
			else {
				/*{{{  a single element in a collapsed dim; save it in temp*/
				/*--------------------------------------------------------------*/
				/* A single element in a collapsed dimension.			*/
				/* sectp = (i, -1,0) shouldn't be changed.  However, we save	*/
				/* it in a temporary scalar (both for dependence and		*/
				/* optimisation purposes).					*/
				/*--------------------------------------------------------------*/
				(*lp)->nlinks--;		/* to 'detach' it */
				attach_expr (*lp, save_scalar_in_temp (*lp));
				/*}}}*/
			}
		}
		
		sectp->attrs |= SE_LOCALISED_SECT;
		/*}}}*/
	}
}
/*}}}*/

/*--------------------------------------*/
/* Special cases of 'attain_mapping ()'	*/
/*--------------------------------------*/
/*{{{  Expr  *broadcast_expr (ep, result_flags)*/
Expr  *broadcast_expr (ep, result_flags)
/*------------------------------------*/
	Expr  *ep;		/* a general expr (which may be array-valued)*/
	int    result_flags;	/* e.g. RESULT_VAR, RESULT_MAPVEC */
/*----------------------------------------------------------------
	This function is called to ensure that an expr (which may
be array-valued) is globally replicated.  If not this function
broadcasts it and overwrites 'ep' with the result.
	It returns a pointer to the resulting expr, which is
always argument 'ep' itself.  It's only provided for ease of use,
e.g. so this can be called within a call to 'attach_expr ()'.
	(Currently this is only called by 'transform_RESHAPE ()',
with 'ep' a vector-valued argument.)

N.B.:  This function is called *instead of* 'attain_mapping (ep)',
so it must call the latter function to evaluate subscripts, actual
args, etc, and to install the mapping 'ep->mapp / sectp'.
----------------------------------------------------------------*/
{
	Map   *tgt_mapp;
	Sect  *tgt_sectp;

	if (ep->type & EX_SCALAR) {
		attain_mapping (ep, NULL_Map, NULL_Sect, result_flags);
		return  ep;
	}

	/*----------------------------------------------*/
	/* ...else 'ep' is an array-valued expr.	*/
	/*----------------------------------------------*/
	choose_map_for_expr (&tgt_mapp, &tgt_sectp, ep);

	replicate_map (tgt_mapp);	/* ...modifies 'tgt_mapp';
					   'tgt_sectp' needn't be changed */

	attain_mapping (ep, tgt_mapp, tgt_sectp, result_flags);
	return  ep;
}
/*}}}*/
/*{{{  Expr  *collapse_expr_on_p0 (ep, result_flags)*/
Expr  *collapse_expr_on_p0 (ep, result_flags)
/*-----------------------------------------*/
	Expr  *ep;		/* a general expr (which may be array-valued)*/
	int    result_flags;	/* e.g. RESULT_VAR, COPY_BACK_IF_VAR */
/*----------------------------------------------------------------
	This function is called by 'transform_and_write_IO_stmt ()'
(directly or indirectly) to ensure that an expr (which may
be array-valued) is *wholly available* on target processor "$p0"
(a scalar processor arrangement).  If not this function
remaps it and overwrites 'ep' with the result.
	It returns a pointer to the resulting expr, which is
always argument 'ep' itself.  It's only provided for ease of use,
e.g. so this can be called within a call to 'attach_expr ()'.

N.B.:
-----
--  This function is called *instead of* 'attain_mapping (ep)',
so it must call the latter function to evaluate subscripts, actual
args, etc, and to install the mapping 'ep->mapp / sectp'.

-- THIS IS ONLY USED TO TRANSFORM I/O STMTS.  *** IT SHOULD BECOME
REDUNDANT WHEN DISTRIBUTED I/O ROUTINES ARE AVAILABLE IN ADLIB ***.
----------------------------------------------------------------*/
{
	Map   *tgt_mapp;
	Sect  *tgt_sectp;

	/*----------------------------------------------*/
	/* Find the required target mapping for 'ep'	*/
	/*----------------------------------------------*/
	if (ep->type & EX_SCALAR) {
		tgt_mapp  = scalar_map_on_p0 ();
		tgt_sectp = NULL_Sect;
	}
	else {
		choose_map_for_expr (&tgt_mapp, &tgt_sectp, ep);
		collapse_map_on_p0 (tgt_mapp);	/* ...modifies 'tgt_mapp';
						   'tgt_sectp' needn't be changed */
	}
	attain_mapping (ep, tgt_mapp, tgt_sectp, result_flags);
	return  ep;
}
/*}}}*/

/*}}}*/

/*{{{  << mapping inquiry functions >>*/
/*---------------------------*/
/* mapping inquiry functions */
/*---------------------------*/
/*{{{  << COMMENT -- mapping formulae >>*/
#if 0
COMMENT -- mapping formulae
---------------------------

For BLOCK
---------
tgt_s  [= the section stride in terms of target subscripts]
        = (s * as)

if ('tgt_s' is compile time const -- so we know whether it's +ve or -ve) {

	loc_l  [= 'l' in terms of local subscripts w.r.t. target array.  Must]
	       [  divide it by |as| to get it in terms of local subscripts.) ]
	       [  N.B. if (loc_l <  0), 'l' is on p < this_p;                ]
	       [       if (loc_l >= b), 'l' is on p > this_p.                ]

	      = al + (l * as) - (b * p)

	eta  [= # of 1st element of the section in the local segment, or,   ]
	     [  more precisely:                                             ]
	     [        if (tgt_s > 0), 1st element >= lb;                    ]
	     [        if (tgt_s < 0), 1st element <= ub.                    ]
	     [ N.B.                                                         ]
	     [(i) if (eta >= n), then this element isn't in the section     ]
	     [        (it's beyond the end of the section).                 ]
	     [(ii) if (eta == 0) then the section may still lie outside the ]
	     [        local segment (i.e. beyond the end of the segment);   ]
	     [(iii) even if (0 <= eta < n), element 'eta' may still not be  ]
	     [        in the local segment if it happens that there are     ]
	     [        none in the local segment at all (because tgt_s > b). ]

	    = MAX ((tgt_s - loc_l - 1) / tgt_s, 0)		! tgt_s > 0
	    = MAX ((tgt_s - loc_l + b) / tgt_s, 0)		! tgt_s < 0

	seg_l  [= local subscript w.r.t. the target array of element 'eta'. ]
	       [     If it's outside the local segment for reasons (ii) or  ]
	       [     (iii) above, then  (seg_l > b-1) if (tgt_s > 0) or     ]
	       [     (seg_l < 0) if (tgt_s < 0).                            ]
	
	      = loc_l + (eta * tgt_s)

<N.B. For the return value, i.e. the actual local subscript,
we must divide this by |as|:      seg_l = seg_l / |as| >

	seg_u = MIN (loc_l + (n-1)*tgt_s,  b-1)			! tgt_s > 0
	      = MAX (loc_l + (n-1)*tgt_s,  0)			! tgt_s < 0

	seg_n = MAX ((seg_u - seg_l + tgt_s)/tgt_s, 0)
}
else {
	use runtime lib fns to compute the above!
}

#endif
/*}}}*/

/*{{{  Expr  *seg_extent (sp, mapp, dim) -- ok *** DOESN'T HANDLE INHERITED TEMPARRAYS! ****/
Expr  *seg_extent (sp, mapp, dim)
/*-----------------------------*/
	Symbol  *sp;	/* used for INHERIT!  What about 'ND_TEMPARRAY' ?? */
	Map     *mapp;
	int      dim;

/*** BY REQUIRING 'sp->name', DOESN'T CATER FOR INHERITED 'TEMPARRAY'S!! ***/
{
	int     array_rank;
	short   fmt;
	Expr   *segn;

	switch (fmt = mapp->fmt [dim]) {
	case  DFMT_COLLAPSED :
		/*------------------------------------------*/
		/* COLLAPSED:  extent = b    (N.B. as == 1) */
		/*------------------------------------------*/
		return  mapp->b [dim];

	case  DFMT_BLOCK :
	case  DFMT_CYCLIC :
		/*------------------------------------------------------*/
		/* extent = { CEILING (b / |as|)            ! BLOCK(b)	*/
		/*          { ncycles * CEILING (b / |as|)  ! CYCLIC(b)	*/
		/*------------------------------------------------------*/
		segn = ceiling_div (mapp->b [dim], abs_func (mapp->as [dim]));
		if (fmt == DFMT_CYCLIC)
			segn = expr (ncycles (mapp, dim), ND_TIMES_OP, segn);

		return  segn;

	case  DFMT_INHERIT :
		array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);

		return  map_primary (sp->name, 0, map_subscr_segn (array_rank, dim));
	}
}
/*}}}*/
/*{{{  Expr  *seg_totlen (sp, mapp)      -- ok ***   "      "       "      " *****/
Expr  *seg_totlen (sp, mapp)
/*------------------------*/
	Symbol  *sp;	/* used for INHERIT!  What about 'ND_TEMPARRAY' ?? */
	Map     *mapp;
{
	int    array_rank,  dim;
	Expr  *len;

	array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	if ( ! array_rank)
		return  int_lit_1;

	len = seg_extent (sp, mapp, 0);		/* i.e. size (dim == 0) */

	for (dim = 1;   dim < array_rank;   ++dim)
		len = expr (len, ND_TIMES_OP, seg_extent (sp, mapp, dim));

	return  len;
}

/*}}}*/
/*{{{  Expr  *ncycles (mapp, dim)        -- OK ***   "      "       "      " *****/
Expr  *ncycles (mapp, dim)
/*----------------------*/
	Map  *mapp;
	int   dim;

/**** DOESN'T CATER FOR 'INHERITED PROCS' AS REFERENCES mapp->proc_sp */
/*----------------------------------------------------------------
Returns an expr for 'ncycles' for a CYCLIC distribution, viz:

	ncycles = ceiling ((1 + al + (an-1)*as) / (b * np))	! as > 0
	        = ceiling ((1 + al) / (b * np))			! as < 0

If 'as' is not a compile-time constant (so it's sign is unknown)
the 'MAX' function is used to return a closed form of the expr:

	ncycles = ceiling ((1 + al + MAX ((an-1)*as, 0)) / (b * np))
----------------------------------------------------------------*/
{
	Expr  *as,  *denom,  *numer;
	int  procdim;

	numer = expr (mapp->al [dim], ND_PLUS_OP, int_lit_1);
	as    = mapp->as [dim];
	if ( ! (as->type & EX_INT_LIT_CONST))
		numer = expr (numer,
                              ND_PLUS_OP,
		              intern_func_ref ("MAX", 2,
		                    expr (expr (mapp->an [dim],
		                                ND_MINUS_OP,
		                                int_lit_1),
		                          ND_TIMES_OP,
		                          as),
		                    int_lit_0)
		             );
	else if (int_val (as) > 0)
		numer = expr (numer,
       	                      ND_PLUS_OP,
		              expr (expr (mapp->an [dim],
		                          ND_MINUS_OP,
		                          int_lit_1),
		                    ND_TIMES_OP,
		                    as)
		             );
	
	procdim = mapp->d [dim];
	denom   = expr (mapp->b [dim],
	                ND_TIMES_OP,
	                mapp->proc_sp->arp->n [procdim]);

	return  ceiling_div (numer, denom);
}
/*}}}*/

/*{{{  int  proc_dim (fmt, array_dim) -- OK*/
int  proc_dim (fmt, array_dim)
/*--------------------------*/
	short  *fmt;		/* i.e. 'fmt []' */
	int     array_dim;	/* >= 0 */
/*----------------------------------------------------------------
	This is called by 'install_id_map ()'.  It returns
the number of the processor array dimension over which dimension
'array_dim' is distributed, according to the distribution format
vector 'fmt []'.  This is given by:

	# (BLOCK / CYCLIC entries in fmt [0 : array_dim - 1])

	If the distribution format is inherited (i.e. "DISTRIBUTE *")
then every element of 'fmt []' contains 'DFMT_INHERIT'.  In this
case the special value -2 is returned, meaning "processor dimension
unknown as it's inherited!"
----------------------------------------------------------------*/
{
	int   dim,  n_distr_dims;

	if (*fmt & DFMT_INHERIT)
		return  -2;	/* special value meaning
				   "unknown as inherited!" */

	n_distr_dims = 0;
	for (dim = array_dim;   dim;   --dim,  ++fmt)
		if (*fmt & (DFMT_BLOCK | DFMT_CYCLIC))
			n_distr_dims++;

	return  n_distr_dims;
}
/*}}}*/

/*{{{  localise_reg_sect_dim (sect_ep,  dim,  sect_dim,  eta_p) -- OK*/
localise_reg_sect_dim (sect_ep,  dim,  sect_dim,  eta_p)
/*----------------------------------------------------*/
	Expr  *sect_ep;		/* ND_ARRAY_SECT, ND_WHOLE_ARRAY or
				   ND_TEMP_ARRAY.  'sect_ep->mapp' & 'sectp'
				   are defined. */
	int    dim,		/* ...a *distributed* dimension */
	       sect_dim;	/* ...the number of previous reg sect dims */
	Expr  **eta_p;		/* if (*eta_p == NULL), it's defined here */
/*{{{  DESCRIPTION*/
#if 0
/*----------------------------------------------------------------
	This is called by 'localise_array_expr ()' to evaluate the
following parameters for a *distributed* regular section dimension,
'dim', of an array reference 'sect_ep':
	(eta, loc_l, loc_n, loc_s)  -- if (*eta_p == NULL) on input, or
	(loc_l, loc_s)              -- if (*eta_p != NULL) on input;

If (*eta_p != NULL) on input, then 'eta' is an input parameter and
the value of 'locn' isn't required.  (This is the situation for
a term on the RHS of an assignment.)   Otherwise both 'eta' and 'locn'
are evaluated here.  '(loc_l /_n /_s)' overwrite 'sectp->(l,n,s)'.

The following formulae apply for BLOCK *or* CYCLIC:

loc_s
-----
= the section stride in terms of target subscripts.
= (s * as)

if ('loc_s' is compile time const -- so we know whether it's +ve or -ve) {

/*{{{  loc_sect_l  (save it in scalar temp)*/
loc_sect_l
----------
= 'l' in terms of local subscripts w.r.t. target array.	 We must
divide it by |as| to get it in terms of local subscripts.

loc_sect_l = al + (l * as) - (b * this_p)


N.B.
-- if (loc_sect_l <  0), 'l' is on p < this_p;
   if (loc_sect_l >= b), 'l' is on p > this_p.

-- For CYCLIC:
	this_p -> this_p + cycle * nprocs
in the above formula.
/*}}}*/
/*{{{  eta      (only compute if undefined on i/p)*/
eta
---
= # of 1st element of the section in the local segment, or more precisely:
        if (loc_s > 0), 1st element >= lb;
        if (loc_s < 0), 1st element <= ub.


eta = MAX ((loc_s - loc_sect_l - 1) / loc_s, 0)		! loc_s > 0
    = MAX ((loc_s - loc_sect_l + b) / loc_s, 0)		! loc_s < 0


N.B.
(i)   only compute this if it's undefined on input.
(ii)  if (eta >= n), then this element isn't in the section (it's
	beyond the end of the section).
(iii) if (eta == 0) then the section may still lie outside the
	local segment (i.e. beyond the end of the segment);
(iv)  even if (0 <= eta < n), element 'eta' may still not be
	in the local segment if it happens that there are none
	in the local segment at all (because loc_s > b).
	

/*}}}*/
/*{{{  loc_l    (save it in scalar temp)*/
loc_l
-----
= local subscript w.r.t. the target array of element 'eta'.
If it's outside the local segment for reasons (ii) or (iii) above,
then  (loc_l > b-1) if (loc_s > 0) or (loc_l < 0) if (loc_s < 0).
		

loc_l = loc_sect_l + (eta * loc_s)


N.B.
For the return value, i.e. the actual local subscript, we must:
(i)  divide this by |as|:

	loc_l -> loc_l / |as|.

(ii) for CYCLIC distribution, add  (cycle * CEILING (b / |as|)):

	locl -> locl + cycle * CEILING (b / |as|)
/*}}}*/
/*{{{  loc_n    (only compute if 'eta' undefined on i/p; save in scalar temp)*/
loc_n
-----
loc_n = MAX ((loc_u - loc_l + loc_s)/loc_s, 0)

where:

loc_u = MIN (loc_sect_l + (n-1)*loc_s,  b-1)		! loc_s > 0
      = MAX (loc_sect_l + (n-1)*loc_s,  0)		! loc_s < 0


N.B.
-- only compute this if it's undefined on input.
/*}}}*/

else {
	use runtime lib fns to compute the above!
}

Save these in scalar temp vars:

loc_sect_l	-- as it's used in the evalaution of all the other parameters
loc_l		} since these are used as DO loop bounds (locn) or in
loc_n		} subscript exprs (locl, locs), and hence are referenced
loc_s		} multiple times.  It also avoids dependence problems.

Need:  l, n, s, al, as, b, p
----------------------------------------------------------------*/
#endif
/*}}}*/

/************************************************
***  FIX THESE LATER! ****
-- 'this_proc_num ()' DOESN'T HANDLE 'INHERIT_PROCS' (cos proc_sp == NULL)!
-- doesn't handle INHERITed distribution format!
************************************************/
{
	/*{{{  decls*/
	bool   calc_eta_and_loc_n,  const_s,  cyclic_distrib,  max_func_in_eta;
	int    s_val;
	
	Expr  *al,  *as,  *b,  *cycle_ep,  *eta,  *ep,  *l,  *loc_l,  *loc_n,
	      *loc_s,  *loc_sect_l,  **l_p,  *n,  **n_p,  *s,  **s_p,  *this_p_ep;
	
	Map   *mapp  = sect_ep->mapp;		/* != NULL */
	Sect  *sectp = sect_ep->sectp;		/* != NULL*/
	/*}}}*/

	calc_eta_and_loc_n = (! *eta_p);
	cyclic_distrib = (mapp->fmt [dim] == DFMT_CYCLIC);
						/* ...else BLOCK or INHERIT */

	/*{{{  initialise pointers l_p, n_p, s_p, l, n, s, al, as, b, eta*/
	l_p = sectp->l + dim;		/* ptr to 'l' */
	n_p = sectp->n + dim;		/* ptr to 'n' */
	s_p = sectp->s + dim;		/* ptr to 's' */
	l   = *l_p; 			/* 'protected' (i.e. nlinks > 0) */
	n   = *n_p;			/*      "        "        "      */
	s   = *s_p;			/*      "        "        "      */
	
	al  = mapp->al [dim];		/* 'protected' (i.e. nlinks > 0) */
	as  = mapp->as [dim];		/*      "        "        "      */
	b   = mapp->b  [dim];		/*      "        "        "      */
	
	eta = *eta_p;
	/*}}}*/
	/*{{{  calc 'this_p_ep'*/
	/*--------------------------------------------------------------*/
	/* Create an expr 'this_p_ep' for the coordinate of the local	*/
	/* processor.							*/
	/*--------------------------------------------------------------*/
	this_p_ep = proc_coord_primary (mapp->proc_sp, mapp->d [dim]);
	
	if (cyclic_distrib) {
		/*----------------------------------------------*/
		/* CYCLIC distribution.  Replace 'p' by:	*/
		/*	p -> p + cycle * n_procs		*/
		/* where:					*/
		/*	cycle = "K_C%d"				*/
		/*----------------------------------------------*/
		char  *str;
	
		str = (char *) emalloc (5);	/* space for, e.g. "K_C0" */
		sprintf (str, "K_C%d", sect_dim);
		cycle_ep = name_primary (str, (EX_INTEGER | EX_SCALAR));
	
		protect_expr (cycle_ep);       /* 'protect' it as it's used again later*/
	
		this_p_ep = expr (this_p_ep,
		                  ND_PLUS_OP,
		                  expr (cycle_ep,
		                        ND_TIMES_OP,
		                        mapp->proc_sp->arp->n [mapp->d [dim]])
		                 );
	}
	/*}}}*/

	attach_expr (loc_s, expr (s, ND_TIMES_OP, as));
	const_s = (loc_s->type & EX_INT_LIT_CONST);

	if (! const_s  &&  calc_eta_and_loc_n) {
		/*{{{  call an 'ADLIB' proc to evaluate params, and return*/
		/*----------------------------------------------------------------
			'loc_s' isn't a compile-time constant, so the values of
		(eta, loc_n) can't be written as F90 expressions as they're
		conditional  (i.e. they involve 'IF's!).  Hence call an 'ADLIB'
		procedure to evaluate them (and 'loc_l'), viz:
		
			CALL AD_LOC_SECT (eta, loc_l, loc_n, mapvec, dim, this_p, cycle)
		
		where the result parameters (eta, loc_l, loc_n) are temp scalar vars.
		----------------------------------------------------------------*/
		Expr  *mapvec_ep;	/* mapvec:  "K_Mname" or "K_TMP (K_M%d)"*/
		
		mapvec_ep = ((sect_ep->token != ND_TEMP_ARRAY) ?
		                 sect_ep->val.sp->mapp->mapvec_ep :	/* K_Mname */
		                 map_primary (NULL_str,  sect_ep->val.allocp->num, -1));
									/* K_TMP (K_M%d) */
		if ( ! cyclic_distrib)
			attach_expr (cycle_ep, int_lit_0);
		
		
		delete_expr (l, DECR_NLINKS);
		delete_expr (n, DECR_NLINKS);
		
		attach_expr (*eta_p, alloc_var (TY_INTEGER, NULL_Map, NULL_Sect));
		attach_expr (*l_p,   alloc_var (TY_INTEGER, NULL_Map, NULL_Sect));
		attach_expr (*n_p,   alloc_var (TY_INTEGER, NULL_Map, NULL_Sect));
		
		stmt ("CALL AD_LOC_SECT (%e, %e, %e, %e, %d, %e, %e)",
		       *eta_p, *l_p, *n_p, mapvec_ep, dim, this_p_ep, cycle_ep);
		
		delete_expr (cycle_ep,  DECR_NLINKS);
		delete_expr (this_p_ep, LEAVE_NLINKS);
		delete_expr (mapvec_ep, LEAVE_NLINKS);
		
		
		/*{{{  finally, loc_s -> loc_s / |as|,  as usual*/
		/*----------------------------------------------------------------
		Now obtain the final value of 'loc_s' as usual, viz:
		
			locs ->   locs / |as|
		N.B.:
		-----
		We delete the old 'sectp->s [dim]' *after* attaching the new expr,
		in case the new expr is exactly the same  (i.e. occupies the same
		storage) as the old one, in which case we don't want to delete it
		before attaching it!
		----------------------------------------------------------------*/
		loc_s->nlinks--;	/* even if (nlinks becomes 0), *don't* delete it! */
		
		if ( ! (as->type & EX_INT_LIT_CONST)  ||  (int_val (as) != 1) )
			loc_s = no_errmsg_div (loc_s, abs_func (as));
		
		/*----------------------------------------------*/
		/* ... else (as == 1), so 'loc_s' is unchanged	*/
		/*----------------------------------------------*/
		
		attach_expr (*s_p, save_scalar_in_temp (loc_s));
		
		delete_expr (s, DECR_NLINKS);
		/*}}}*/
		
		
		return;
		/*}}}*/
	}

	if (const_s  &&  ! (s_val = int_val (loc_s)) ) {
		/*{{{  (loc_s == 0) is an error;  substitute 1 as a simple fix!*/
		/*--------------------------------------------------------------*/
		/* (loc_s == 0) is an error that would cause divide-by-0.	*/
		/* Replace it by 1 as a simple fix!				*/
		/*--------------------------------------------------------------*/
		unprotect_expr (loc_s);			/* i.e. delete it */
		
		attach_expr (loc_s, int_lit_1);
		s_val = 1;
		/*}}}*/
	}

	/*{{{  calc 'loc_sect_l' (and save it in scalar temp) -- OK*/
	/*----------------------------------------------------------------
	loc_sect_l = al + (l * as) - (b * this_p)
	
	This is the subscript of the 1st element of the regular section
	(whose *parent* subscript is 'l'), in terms of *local* subscripts
	-- but without dividing by |as|!
	
	N.B.:
	-- For CYCLIC distribution, 'this_p' is actually:
	
		this_p + "K_C%d" * n_procs
	
	(where 'K_C%d' is a DO-variable of an enclosing DO-loop).
	-- We assign this expr to a temporary variable (and output this
	assignment) for efficiency, since it appears explicitly in the
	formulae for all of (eta, loc_l, loc_n).
	----------------------------------------------------------------*/
	attach_expr (loc_sect_l,
	             save_scalar_in_temp (
	                           expr (al,
	                                 ND_PLUS_OP,
	                                 expr (expr (l, ND_TIMES_OP, as),
	                                       ND_MINUS_OP,
	                                       expr (b, ND_TIMES_OP, this_p_ep)
	                                      )
	                                )
	             ));
	/*}}}*/
	if (calc_eta_and_loc_n) {
		/*{{{  calc 'eta' -- OK*/
		/*----------------------------------------------------------------
		eta: the number of the 1st element (if there are any) of the
		regular section that is local to this processor, i.e., whose
		subscript is:
			> lb (local segment)   -- loc_s > 0
			< ub (local segment)   -- loc_s < 0
		
		if (loc_s > 0)
			eta = MAX (CEILING (-loc_sect_l / loc_s), 0)
			    = MAX ((loc_s - loc_sect_l - 1) / loc_s, 0)
		
		else
			eta = MAX (CEILING ((loc_sect_l - loc_ub) / (-loc_s)), 0)
			    = MAX (CEILING ((loc_sect_l - b + 1) / (-loc_s)), 0)
			    = MAX ((loc_s - loc_sect_l + b) / loc_s, 0)
		
		N.B.:
		-----
		(1) This can't be evaluated statically as it involves 'loc_sect_l'
		-- a non-constant.
		
		(2)  If (loc_s > 0) and (l == al == 0), then eta is guaranteed
		to be >= 0, so we can omit the "MAX (..,0)" ref.
		----------------------------------------------------------------*/
		
		ep = expr (loc_s, ND_MINUS_OP, loc_sect_l);
		
		ep = (s_val > 0) ? expr (ep, ND_MINUS_OP, int_lit_1) :
		                   expr (ep, ND_PLUS_OP, b);
		
		eta = expr (ep, ND_DIV_OP, loc_s);
		
		/*--------------------------------------------------------------*/
		/* If the following test is FALSE then:				*/
		/*	(- loc_sect_l)  =  (b * this_p)  >=  0			*/
		/*	        loc_s   >  0					*/
		/* so it's guaranteed that (eta >= 0), so no 'MAX' function	*/
		/* reference is required!					*/
		/*--------------------------------------------------------------*/
		max_func_in_eta = ((s_val < 0)  ||
		                 ! (l->type & EX_INT_LIT_CONST)  ||  int_val (l)  ||
		                 ! (al->type & EX_INT_LIT_CONST) ||  int_val (al));
		
		if (max_func_in_eta)
			eta = intern_func_ref ("MAX", 2, eta, int_lit_0);
		
		attach_expr (*eta_p, eta);
		/*}}}*/
	}
	else
		max_func_in_eta = ((eta->token == ND_INTERN_FUNC)  &&
		                    str_match (eta->val.strval, "MAX"));
	/*{{{  calc 'loc_l'*/
	/*----------------------------------------------------------------
	loc_l = loc_sect_l + (eta * loc_s)
	
	This is the subscript of the 1st *local* element of the regular
	section (if there are any), expressed in *local* subscripts.
	
	Special simple cases are:
	
	if (s_val == 1)
		loc_l = loc_sect_l + eta
		      = loc_sect_l + MAX (- loc_sect_l, 0)
		      = MAX (loc_sect_l, 0)
	
	if (s_val == 1  &&  ! max_func_in_eta)
		loc_l = 0
	
	N.B.:
	-----
	(1) When (eta == 0), it may happen that:
	
	       (loc_l > loc_ub)		-- (loc_s > 0)
	   or  (loc_l < 0)		-- (loc_s < 0)
	
	in which case the regular section lies totally outside the local
	segment (i.e. there are *no* local elements, so we should get
	(loc_n == 0)!).
	
	(2) It can be seen that:
		loc_l >= 0		-- (loc_s > 0)
		loc_l <= loc_ub		-- (loc_s < 0)
	----------------------------------------------------------------*/
	if (const_s  &&  (s_val == 1))
		loc_l = max_func_in_eta ?
		           intern_func_ref ("MAX", 2, loc_sect_l, int_lit_0) :
		           int_lit_0;
	else
		loc_l = expr (loc_sect_l, ND_PLUS_OP, expr (eta, ND_TIMES_OP, loc_s));
	
	protect_expr (loc_l);
	/*}}}*/
	if (calc_eta_and_loc_n) {
		/*{{{  calc 'loc_n' (and save it in scalar temp) -- OK*/
		/*----------------------------------------------------------------
		loc_n = (loc_u - loc_l + loc_s) / loc_s
		
		where:
		
		loc_u = MIN (loc_sect_l + (n-1)*loc_s,  b-1)	-- loc_s > 0
		      = MAX (loc_sect_l + (n-1)*loc_s,  0)	-- loc_s < 0
		
		Really the RHS expr should be contained in a 'MAX' func ref:
		
			loc_n = MAX (..., 0)
		
		but we omit it for reasons given below.
			Special simple cases are:
		
		if (s_val == 1)
			loc_n = loc_u - loc_l + 1
			      = MIN (loc_sect_l + n-1,  b-1) - loc_l + 1
			      = MIN (loc_sect_l + n,  b) - loc_l
		
		if (s_val == 1  &&  ! max_func_in_eta)
			...this automatically simplifies further, since (loc_l == 0)
		
		N.B.:
		-----
		-- This can't be evaluated statically as 'loc_u' and 'loc_l'
		involve 'loc_sect_l', which depend on 'this_p'.
		
		-- Since this is used in the context:
		
			DO i = 0, loc_n - 1
		
		we omit the 'MAX' function ref in the definition of 'loc_n'.
		If the 1st argument is <= 0, then whatever it's value, it is
		equivalent to 0 (in that the DO-loop count is 0).
		
		--  We delete the old 'sectp->n [dim]' *after* attaching the
		new one, in case the new expr is exactly the same (i.e. occupies
		the same storage) as the old one, in which case we don't want to
		delete it before attaching it!
		----------------------------------------------------------------*/
		if (s_val == 1)
			loc_n = expr (intern_func_ref ("MIN", 2,
			                                expr (loc_sect_l, ND_PLUS_OP, n),
			                                b),
			              ND_MINUS_OP,
			              loc_l);
		else {
			/*{{{  use the full formula for 'loc_n'*/
			/*--------------------------------------------------------------*/
			/* loc_n = (loc_u - loc_l + loc_s) / loc_s			*/
			/*								*/
			/* where:							*/
			/*								*/
			/* loc_u = MIN (loc_sect_l + (n-1)*loc_s,  b-1)	-- loc_s > 0	*/
			/*       = MAX (loc_sect_l + (n-1)*loc_s,  0)	-- loc_s < 0	*/
			/*--------------------------------------------------------------*/
			Expr  *loc_u;
			
			ep = expr (loc_sect_l,
			           ND_PLUS_OP,
			           expr (expr (n, ND_MINUS_OP, int_lit_1),
			                 ND_TIMES_OP,
			                 loc_s)
			          );
			
			loc_u = (s_val > 0) ? intern_func_ref ("MIN", 2, ep,
			                                        expr (b, ND_MINUS_OP, int_lit_1)) :
			                      intern_func_ref ("MAX", 2, ep, int_lit_0);
			
			loc_n = expr (expr (expr (loc_u, ND_MINUS_OP, loc_l),
			                    ND_PLUS_OP,
			                    loc_s),
			              ND_DIV_OP,
			              loc_s);
			/*}}}*/
		}
		attach_expr (*n_p,  save_scalar_in_temp (loc_n));
		delete_expr (n, DECR_NLINKS);
		/*}}}*/
	}
	
	/*{{{  convert (loc_l, loc_s) to local subscript values (& save in temps)*/
	/*----------------------------------------------------------------
	For the final values of 'loc_l' and 'loc_s'  (i.e. the *actual*
	local subscripts)  we must change them as follows:
	
		locl -> { locl / |as|			! BLOCK distrib
		        { locl / |as| + "K_C%d" * CEILING (b / |as|)
							! CYCLIC distrib
		locs ->   locs / |as|
		        
	N.B.:
	-----
	We delete the old 'sectp->l [dim]' & 's [dim]' *after* attaching
	the new exprs, in case the new expr is exactly the same  (i.e.
	occupies the same storage) as the old one, in which case we don't
	want to delete it before attaching it!
	----------------------------------------------------------------*/
	loc_l->nlinks--;	/* even if (nlinks becomes 0), *don't* delete it! */
	loc_s->nlinks--;	/*    "      "       "      "       "      "      */
	
	if ( ! (as->type & EX_INT_LIT_CONST)  ||  (int_val (as) != 1) ) {
		loc_l = no_errmsg_div (loc_l, abs_func (as));
		loc_s = no_errmsg_div (loc_s, abs_func (as));
	}
	/*--------------------------------------------------------------*/
	/* ... else (as == 1), so 'loc_l' and 'loc_s' are unchanged	*/
	/*--------------------------------------------------------------*/
	
	if (cyclic_distrib) {
		/*------------------------------------------------------*/
		/* CYCLIC distribution, so replace 'locl' by:		*/
		/*	locl -> locl + "K_C%d" * CEILING (b / |as|)	*/
		/*------------------------------------------------------*/
		loc_l = expr (loc_l,
		              ND_PLUS_OP,
		              expr (cycle_ep,
		                    ND_TIMES_OP,
		                    ceiling_div (b, abs_func (as)) ) );
	
		cycle_ep->nlinks--;	/* To 'undo' earlier protection.  Even
					   if (nlinks -> 0), *don't* delete it! */
	}
	attach_expr (*l_p, save_scalar_in_temp (loc_l));
	attach_expr (*s_p, save_scalar_in_temp (loc_s));
	
	delete_expr (l, DECR_NLINKS);
	delete_expr (s, DECR_NLINKS);
	/*}}}*/

	unprotect_expr (loc_sect_l);	/* delete it if unused */
}
/*}}}*/

/*{{{  proc_and_subscript (proc_p,  loc_i_p,  mapp,  parent_i,  dim)*/
proc_and_subscript (proc_p,  loc_i_p,  mapp,  parent_i,  dim)
/*---------------------------------------------------------*/
	Expr  **proc_p;		/* if != NULL, address for 'proc' result */
	Expr  **loc_i_p;	/* if != NULL, address for 'loc_i' result */
	Map    *mapp;		/* != NULL */
	Expr   *parent_i;	/* subscript w.r.t. parent array in
				                     dimension 'dim' */
	int     dim;		/* ...a *distributed* array dim of 'mapp' */
/*----------------------------------------------------------------
	Calculates the subscript of the owner processor, 'owner_proc',
and/or the local subscript, 'loc_i', corresponding to 'parent'
subscript 'parent_i' in (distributed) dimension 'dim' of mapping
'mapp'.
	If 'proc_p' and 'loc_i_p' are != NULL, the results 'owner_proc'
and 'loc_i' are returned in '*proc_p' and '*loc_i_p' resp.

For BLOCK
---------
tgt_i = al + (parent_i * as)

owner_proc = tgt_i / b
loc_i      = MOD (tgt_i, b) / |as|

For CYCLIC
----------
as above, then (in this order!)...

owner_proc = MOD (owner_proc, n_procs)
loc_i      = loc_i + (owner_proc / n_procs) * ceiling_div (b, |as|)
----------------------------------------------------------------*/
/************************************************
-- Doesn't handle INHERITed distribution format !
************************************************/
{
	bool  cyclic_distrib = (mapp->fmt [dim] == DFMT_CYCLIC);
	Expr  *al,  *as,  *b,  *loc_i,  *owner_proc,  *tgt_i;

	al = mapp->al [dim];	/* this is 'protected' (has nlinks == 1) */
	as = mapp->as [dim];	/*      "       "        "       "       */
	b  = mapp->b [dim];	/*      "       "        "       "       */
	protect_expr (parent_i);

	/*{{{  tgt_i = al + (parent_i * as)*/
	/*------------------------------*/
	/* tgt_i = al + (parent_i * as) */
	/*------------------------------*/
	tgt_i = expr (al, ND_PLUS_OP, expr (parent_i, ND_TIMES_OP, as));
	
	protect_expr (tgt_i);
	/*}}}*/
	/*{{{  evaluate (owner_proc, loc_i)*/
	/*--------------------------------------*/
	/* owner_proc = tgt_i / b		*/
	/* loc_i      = MOD (tgt_i, b) / |as|	*/
	/*--------------------------------------*/
	if (proc_p  ||  cyclic_distrib)
		owner_proc = expr (tgt_i, ND_DIV_OP, b);
	
	if (loc_i_p)
		loc_i = no_errmsg_div (mod_func (tgt_i, b), abs_func (as));
	/*}}}*/

	if (cyclic_distrib) {
		/*{{{  modify (owner_proc, loc_i) for CYCLIC distribution*/
		/*----------------------------------------------------------------
		Modify (owner_proc, loc_i) for CYCLIC distribution as follows:
		
		loc_i      = loc_i + (owner_proc / n_procs) * ceiling_div (b, |as|)
		owner_proc = MOD (owner_proc, n_procs)
		
		(evaluated in this order!!)
		----------------------------------------------------------------*/
		Expr  *n_procs,  *own_proc;
		
		n_procs = mapp->proc_sp->arp->n [mapp->d [dim]];
		attach_expr (own_proc, owner_proc);
						/* save pointer to 'owner_proc' so we can
						   delete it if unused, and 'protect' it */
		if (loc_i_p)
			loc_i = expr (loc_i,
			              ND_PLUS_OP,
			              expr (no_errmsg_div (own_proc, n_procs),
			                    ND_TIMES_OP,
			                    ceiling_div (b, abs_func (as)) ) );
		
		if (proc_p)
			owner_proc = mod_func (own_proc, n_procs);
		
		delete_expr (own_proc, DECR_NLINKS);		/* delete it if unused! */
		/*}}}*/
	}
	unprotect_expr (tgt_i);		/* delete if unused */
	parent_i->nlinks--;		/* to remove earlier 'protection' */

	if (loc_i_p) attach_expr (*loc_i_p, loc_i);
	if (proc_p)  attach_expr (*proc_p,  owner_proc);
}
/*}}}*/

/*{{{  bool  must_call_scatter (var_ep) -- ok, but check*/
bool  must_call_scatter (var_ep)
/*----------------------------*/
	Expr  *var_ep;		/* (should be) an assignment variable.  If
				   it's an array section it has 'sectp' set.*/
/*----------------------------------------------------------------
	Called by 'write_assignment_stmt (var_ep, expr_ep)' to
determine whether 'var_ep' is an irregular section *and* 'AD_SCATTER()'
must be called to assign the result of 'expr_ep' to 'var_ep'.
I.e. if this returns TRUE, then:

	var = expr

must be translated as:

	alloc (temp)
	temp = expr
	CALL AD_SCATTER (var, temp, v1, ...)
	dealloc (temp)
	
	This condition is TRUE if 'var_ep' has a vector subscript
in any *distributed* (rather than collapsed) dimension.
----------------------------------------------------------------*/
{
	int     array_rank,  dim;
	short  *dp;
	Expr  **np;

	if (var_ep->token != ND_ARRAY_SECT)
		return  FALSE;

	array_rank = (var_ep->sectp->attrs & SE_RANK_MASK);	/* > 0 */

	dp = var_ep->mapp->d;
	np = var_ep->sectp->n;
	for (dim = array_rank;   dim;   --dim,  ++np,  ++dp) {
		if ((*np == int_lit_m2)  &&  (*dp != -1))
			/*----------------------------------------
			A vector subscript in a distributed dim.
			Return TRUE.
			----------------------------------------*/
			return  TRUE;
	}

	return  FALSE;
}
/*}}}*/

/*{{{  bool  is_distrib (ep) -- OK*/
bool  is_distrib (ep)
/*-----------------*/
	Expr  *ep;		/* A general expr.  We don't assume that
				   it's had its mapping 'setup' yet! */
/*----------------------------------------------------------------
	Return TRUE if the 'natural' mapping for expr 'ep'
(i.e. that which would be returned by 'choose_map_for_expr (ep)')
is distributed or replicated on a named processor arrangement,
and FALSE otherwise (i.e. if the natural mapping is globally
replicated).
	(Currently this is only called by 'transform_DOT_PRODUCT()'
to determine which of its 2 args should bequeath its mapping to
the other.)

N.B.
-- It isn't crucial that this returns the 'correct' answer --
it's only used to try to establish a 'good' mapping.
-- If the 'first' array-valued term is a reference to a
transformational intrinsic function, this will 'setup' its mapping;
otherwise it won't setup the mapping of any terms in 'ep'.
----------------------------------------------------------------*/
{
	Map   *mapp;
	Sect  *sectp;

	if (ep->type & EX_SCALAR)
		return  FALSE;

	ep = find_array_term (ep);	/* ...a 'ND_FUNC /_ARRAY_SECT /
					   _WHOLE_ARRAY /_TEMP_ARRAY' term*/
	if (ep->mapp)
		/*----------------------------------------------*/
		/* 'ep's mapping has already been set-up (which	*/
		/* includes the case where it's a temp array).	*/
		/*----------------------------------------------*/
		mapp = ep->mapp;

	else if (ep->token != ND_FUNC)
		/*----------------------------------------------*/
		/* 'ep' is a 'ND_WHOLE_ARRAY' or 'ND_ARRAY_SECT'*/
		/*----------------------------------------------*/
		mapp = ep->val.sp->mapp;

	else {
		/*----------------------------------------------*/
		/* 'ep' is an array-valued function (external or*/
		/* transformational intrinsic).  This means	*/
		/* that it must have an explicit interface.	*/
		/*----------------------------------------------*/
		Symbol  *result_sp = ep->val.sp->info.sp;

		if ( ! (result_sp->attrs & AT_INTRINSIC))
			/*----------------------------*/
			/* an external array function */
			/*----------------------------*/
			mapp = result_sp->mapp;
		else
			/*---------------------------------------*/
			/* a transformational intrinsic function! */
			/*---------------------------------------*/
			choose_map_for_expr (&mapp, &sectp, ep);
	}
	return  (mapp  &&  (mapp->attrs & (MA_DISTRIB | MA_REPLIC_ON_PROCS)));
}
/*}}}*/
/*}}}*/

/*{{{  << TESTING FUNCTIONS -- DELETE LATER !! >>*/

/*{{{  testmsg_align (func, sp, alignp)  *** FOR TESTING ONLY -- DELETE LATER ****/
testmsg_align (func, sp, alignp)
/*----------------------------*/
	char    *func;		/* calling function name */
	Symbol  *sp;		/* alignee; may be NULL */
	Align   *alignp;	/* alignp; may have (alignp->tgt_sp == NULL) */
{
	int    i;
	int32  attrs       = alignp->attrs;
	int    src_rank    = (attrs & AL_SRC_RANK_MASK),
	       embed_rank  = (attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT,
	       replic_rank = (attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;

	testmsg (func);
	fputs ("            ", stderr);
	for (i = strlen (func);  i;  --i)
		putc ('-', stderr);
	putc ('\n', stderr);

	testmsg ("alignee = %q;  target = %q\n",
	            (sp ? sp->name : NULL_str),
	            (alignp->tgt_sp ? alignp->tgt_sp->name : NULL_str) );
	            
	testmsg ("i\tal [i]\tas [i]\td [i]\n            -\t------\t------\t-----");
	for (i = 0;   i < src_rank;   i++)
		testmsg ("%d\t%e\t%e\t%d", i, alignp->al[i], alignp->as[i], alignp->d[i]);

	testmsg ("\n            i\tei [i]\t\ted [i]\n            -\t-------\t\t------");
	for (i = 0;   i < embed_rank;   i++)
		testmsg ("%d\t%e\t\t%d", i, alignp->ei[i], alignp->ed[i]);

	testmsg ("\n            i\trd [i]\n            -\t------");
	for (i = 0;   i < replic_rank;   i++)
		testmsg ("%d\t%d", i, alignp->rd[i]);
}
/*}}}*/

/*{{{  testmsg_distr (func, sp, distp)  *** FOR TESTING ONLY -- DELETE LATER ****/
testmsg_distr (func, sp, distp)
/*----------------------------*/
	char    *func;		/* calling function name */
	Symbol  *sp;		/* distributee; may be NULL */
	Distr   *distp;
{
	int    i;
	int32  attrs      = distp->attrs;
	int    src_rank   = (attrs & DI_SRC_RANK_MASK),
	       proc_rank  = (attrs & DI_PROC_RANK_MASK) >> DI_PROC_RANK_SHIFT;
	       
	testmsg (func);
	fputs ("            ", stderr);
	for (i = strlen (func);  i;  --i)
		putc ('-', stderr);
	putc ('\n', stderr);

	testmsg ("src = %q;  target proc = %q;  src_rank = %d;  proc_rank = %d",
	            (sp ? sp->name : NULL_str),
	            (distp->proc_sp ? distp->proc_sp->name : NULL_str),
	            src_rank,  proc_rank);

	testmsg ("%s fmt;    %s procs\n",
	          (attrs & DI_INHERIT_FMT ? "Inherit" :
	          (attrs & DI_ASSERT_FMT ? "Assert" :
	          (src_rank ? "Prescribe" :  "No"))),
	          (attrs & DI_INHERIT_PROCS ? "Inherit" :
	          (attrs & DI_ASSERT_PROCS ? "Assert" :
	          (distp->proc_sp ? "Prescribe" :  "No"))) );

	testmsg ("i\tfmt [i]\tb [i]\n            -\t---\t-");
	for (i = 0;   i < src_rank;   i++) {
		char  *fmt_str;

		switch (distp->fmt [i]) {
		case  DFMT_BLOCK :	fmt_str = "bl";		break;
		case  DFMT_CYCLIC :	fmt_str = "cy";		break;
		case  DFMT_COLLAPSED :	fmt_str = "*";		break;
		case  DFMT_INHERIT :	fmt_str = "inherit";	break;
		}
		if (distp->b [i])
			testmsg ("%d\t%s\t%e", i, fmt_str, distp->b [i]);
		else
			testmsg ("%d\t%s\t--", i, fmt_str);
	}
}
/*}}}*/

/*{{{  testmsg_map (func, sp, mapp)  *FOR TESTING ONLY -- DELETE LATER! ****/
testmsg_map (func, sp, mapp)
/*------------------------*/
	char    *func;		/* calling function name */
	Symbol  *sp;		/* mapped variable; may be NULL */
	Map     *mapp;
{
	int    i;
	int32  attrs       = mapp->attrs;
	int    array_rank  = (attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT,
	       embed_rank  = (attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT;

	testmsg (func);
	fputs ("            ", stderr);
	for (i = strlen (func);  i;  --i)
		putc ('-', stderr);
	putc ('\n', stderr);
	
	testmsg ("var = %q;  target proc = %q;  array_rank = %d;  embed_rank = %d",
	            (sp ? sp->name : NULL_str),
	            (mapp->proc_sp ? mapp->proc_sp->name : NULL_str),
	            array_rank,  embed_rank);

	testmsg ("%s;    %s\n",
	         ((attrs & MA_DISTRIB) ? "Distributed" : "NOT distributed"),
	         ( (attrs & MA_INHERITED) ? "INHERITed" :
	          ((attrs & MA_ALIGNED_WITH_INHERIT) ? "ALIGNed with INHERIT" :
	                                    "NOT INHERITed") ) );
	/*{{{  output (al, an, as, tn, b, fmt, d)*/
	testmsg ("i\tal\tan\tas\ttn\tb\tfmt\td\n           -\t--\t--\t--\t-\t--\t---\t-");
	for (i = 0;   i < array_rank;   i++) {
		char  *fmt_str;
	
		switch (mapp->fmt [i]) {
		case  DFMT_BLOCK :	fmt_str = "bl";		break;
		case  DFMT_CYCLIC :	fmt_str = "cy";		break;
		case  DFMT_COLLAPSED :	fmt_str = "*";		break;
		case  DFMT_INHERIT :	fmt_str = "inherit";	break;
		}
		testmsg ("%d\t%e\t%e\t%e\t%e\t%e\t%s\t%d",
		          i,
		          mapp->al [i],  mapp->an [i],  mapp->as [i],
		          mapp->tn [i],  mapp->b [i],
		          fmt_str,       mapp->d [i]);
	}
	/*}}}*/

	/*{{{  output (ep, ed)*/
	testmsg ("i\tep\ted\n           -\t--\t--");
	for (i = 0;   i < embed_rank;   i++)
		testmsg ("%d\t%e\t%d",  i,   mapp->ep [i],   mapp->ed [i]);
	/*}}}*/
}

/*}}}*/

/*{{{  testmsg_map_and_sect (func, sect_ep, mapp, sectp) * FOR TESTING ONLY -- DELET LATER*/
testmsg_map_and_sect (func, sect_ep, mapp, sectp)
/*---------------------------------------------*/
	char    *func;		/* calling function name */
	Expr    *sect_ep;	/* may be NULL */
	Map     *mapp;		/* may be NULL.  Usually 'sect_ep->mapp' */
	Sect    *sectp;		/* may be NULL.  Usually 'sect_ep->sectp'*/
{
	int      array_rank,  embed_rank,  i;
	int32    attrs;

	attrs       = mapp->attrs;
	array_rank  = (attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT;
	embed_rank  = (attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT;
	
	testmsg (func);
	fputs ("            ", stderr);
	for (i = strlen (func);  i;  --i)
		putc ('-', stderr);
	putc ('\n', stderr);
	
	testmsg ("sect_ep = %e;  target proc = %q;  array_rank = %d;  embed_rank = %d",
	            sect_ep,
	            (mapp->proc_sp ? mapp->proc_sp->name : NULL_str),
	            array_rank,  embed_rank);

	testmsg ("%s;    %s\n",
	         ((attrs & MA_DISTRIB) ? "Distributed" : "NOT distributed"),
	         ( (attrs & MA_INHERITED) ? "INHERITed" :
	          ((attrs & MA_ALIGNED_WITH_INHERIT) ? "ALIGNed with INHERIT" :
	                                    "NOT INHERITed") ) );

	/*{{{  output (al, an, as, tn, b, fmt, d)*/
	testmsg ("i\tal\tan\tas\ttn\tb\tfmt\td\n           -\t--\t--\t--\t-\t--\t---\t-");
	for (i = 0;   i < array_rank;   i++) {
		char  *fmt_str;
	
		switch (mapp->fmt [i]) {
		case  DFMT_BLOCK :	fmt_str = "bl";		break;
		case  DFMT_CYCLIC :	fmt_str = "cy";		break;
		case  DFMT_COLLAPSED :	fmt_str = "*";		break;
		case  DFMT_INHERIT :	fmt_str = "inherit";	break;
		}
		testmsg ("%d\t%e\t%e\t%e\t%e\t%e\t%s\t%d",
		          i,
		          mapp->al [i],  mapp->an [i],  mapp->as [i],
		          mapp->tn [i],  mapp->b [i],
		          fmt_str,       mapp->d [i]);
	}
	/*}}}*/

	/*{{{  output (ep, ed)*/
	testmsg ("i\tep\ted\n           -\t--\t--");
	for (i = 0;   i < embed_rank;   i++)
		testmsg ("%d\t%e\t%d",  i,   mapp->ep [i],   mapp->ed [i]);
	/*}}}*/

	if (sectp) {
		/*{{{  output (l, n, s)*/
		testmsg ("i\tl\tn\ts\n           -\t-\t-\t-");
		
		for (i = 0;   i < array_rank;   i++)
			testmsg ("%d\t%e\t%e\t%e", i, sectp->l [i], sectp->n [i], sectp->s [i]);
		/*}}}*/
	}
	else
		testmsg ("no \'sectp\' parameters (l,n,s) -- they\'re == (0, an, 1)");

}


/*}}}*/
/*}}}*/

#endif
/*}}}*/
